# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import doctest
import importlib
import os.path
import pkg_resources
import random
import re
import sys
import textwrap
import unittest

import trac
from trac import util
from trac.test import mkdtemp, rmtree
from trac.util.tests import (concurrency, datefmt, presentation, text,
                             translation, html)


class AtomicFileTestCase(unittest.TestCase):

    def setUp(self):
        self.dir = mkdtemp()
        self.path = os.path.join(self.dir, 'trac-tempfile')

    def tearDown(self):
        rmtree(self.dir)

    def test_non_existing(self):
        with util.AtomicFile(self.path) as f:
            f.write('test content')
        self.assertTrue(f.closed)
        self.assertEqual('test content', util.read_file(self.path))

    def test_existing(self):
        util.create_file(self.path, 'Some content')
        self.assertEqual('Some content', util.read_file(self.path))
        with util.AtomicFile(self.path) as f:
            f.write('Some new content')
        self.assertTrue(f.closed)
        self.assertEqual('Some new content', util.read_file(self.path))

    if os.name != 'nt':
        def test_symbolic_link(self):
            link_path = os.path.join(self.dir, 'trac-tempfile-link')
            os.symlink(self.path, link_path)

            with util.AtomicFile(link_path) as f:
                f.write('test content')

            self.assertTrue(os.path.islink(link_path))
            self.assertEqual('test content', util.read_file(link_path))
            self.assertEqual('test content', util.read_file(self.path))

    if util.can_rename_open_file:
        def test_existing_open_for_reading(self):
            util.create_file(self.path, 'Initial file content')
            self.assertEqual('Initial file content', util.read_file(self.path))
            with open(self.path) as rf:
                with util.AtomicFile(self.path) as f:
                    f.write('Replaced content')
            self.assertTrue(rf.closed)
            self.assertTrue(f.closed)
            self.assertEqual('Replaced content', util.read_file(self.path))

    # FIXME: It is currently not possible to make this test pass on all
    # platforms and with all locales. Typically, it will fail on Linux with
    # LC_ALL=C.
    # Python 3 adds sys.setfilesystemencoding(), which could be used here
    # to remove the dependency on the locale. So the test is disabled until
    # we require Python 3.
    def _test_unicode_path(self):
        self.path = os.path.join(self.dir, u'träc-témpfilè')
        with util.AtomicFile(self.path) as f:
            f.write('test content')
        self.assertTrue(f.closed)
        self.assertEqual('test content', util.read_file(self.path))


class PathTestCase(unittest.TestCase):

    def assert_below(self, path, parent):
        self.assertTrue(util.is_path_below(path.replace('/', os.sep),
                                           parent.replace('/', os.sep)))

    def assert_not_below(self, path, parent):
        self.assertFalse(util.is_path_below(path.replace('/', os.sep),
                                            parent.replace('/', os.sep)))

    def test_is_path_below(self):
        self.assert_below('/svn/project1', '/svn/project1')
        self.assert_below('/svn/project1/repos', '/svn/project1')
        self.assert_below('/svn/project1/sub/repos', '/svn/project1')
        self.assert_below('/svn/project1/sub/../repos', '/svn/project1')
        self.assert_not_below('/svn/project2/repos', '/svn/project1')
        self.assert_not_below('/svn/project2/sub/repos', '/svn/project1')
        self.assert_not_below('/svn/project1/../project2/repos',
                              '/svn/project1')
        self.assertTrue(util.is_path_below('repos', os.path.join(os.getcwd())))
        self.assertFalse(util.is_path_below('../sub/repos',
                                            os.path.join(os.getcwd())))

    def test_native_path(self):
        self.assertIsNone(util.native_path(None))
        if os.name == 'posix':
            self.assertEqual('/D/Trac/x', util.native_path('D:\\Trac\\x'))
            self.assertEqual('/D/Trac/x', util.native_path('/D/Trac/x'))
            self.assertEqual('/D/', util.native_path('D:\\'))
            self.assertEqual('/Trac/x', util.native_path('\\Trac\\x'))
            self.assertEqual('Trac/x', util.native_path('Trac\\x'))
            self.assertEqual('Trac/x', util.native_path('Trac/x'))
        elif os.name == 'nt':
            self.assertEqual('D:\\Trac\\x', util.native_path('/D/Trac/x'))
            self.assertEqual('D:\\Trac\\x', util.native_path('D:/Trac/x'))
            self.assertEqual('D:\\Trac\\x', util.native_path('D:\\Trac\\x'))
            self.assertEqual('D:\\', util.native_path('/D/'))
            self.assertEqual('D:', util.native_path('/D'))
            self.assertEqual('C:\\', util.native_path('/'))
            self.assertEqual('C:\\Trac\\x', util.native_path('/Trac/x'))
            self.assertEqual('Trac\\x', util.native_path('Trac/x'))
            self.assertEqual('Trac\\x', util.native_path('Trac\\x'))

class RandomTestCase(unittest.TestCase):

    def setUp(self):
        self.state = random.getstate()

    def tearDown(self):
        random.setstate(self.state)

    def test_urandom(self):
        """urandom() returns random bytes"""
        for i in xrange(129):
            self.assertEqual(i, len(util.urandom(i)))
        # For a large enough sample, each value should appear at least once
        entropy = util.urandom(65536)
        values = {ord(c) for c in entropy}
        self.assertEqual(256, len(values))

    def test_hex_entropy(self):
        """hex_entropy() returns random hex digits"""
        hex_digits = set('0123456789abcdef')
        for i in xrange(129):
            entropy = util.hex_entropy(i)
            self.assertEqual(i, len(entropy))
            self.assertEqual(set(), set(entropy) - hex_digits)

    def test_hex_entropy_global_state(self):
        """hex_entropy() not affected by global random generator state"""
        random.seed(0)
        data = util.hex_entropy(64)
        random.seed(0)
        self.assertNotEqual(data, util.hex_entropy(64))


class ContentDispositionTestCase(unittest.TestCase):

    def test_filename(self):
        self.assertEqual('attachment; filename=myfile.txt',
                         util.content_disposition('attachment', 'myfile.txt'))
        self.assertEqual('attachment; filename=a%20file.txt',
                         util.content_disposition('attachment', 'a file.txt'))

    def test_no_filename(self):
        self.assertEqual('inline', util.content_disposition('inline'))
        self.assertEqual('attachment', util.content_disposition('attachment'))

    def test_no_type(self):
        self.assertEqual('filename=myfile.txt',
                         util.content_disposition(filename='myfile.txt'))
        self.assertEqual('filename=a%20file.txt',
                         util.content_disposition(filename='a file.txt'))


class SafeReprTestCase(unittest.TestCase):
    def test_normal_repr(self):
        for x in ([1, 2, 3], "été", u"été"):
            self.assertEqual(repr(x), util.safe_repr(x))

    def test_buggy_repr(self):
        class eh_ix(object):
            def __repr__(self):
                return 1 + "2"
        self.assertRaises(Exception, repr, eh_ix())
        sr = util.safe_repr(eh_ix())
        sr = re.sub('[A-F0-9]{4,}', 'ADDRESS', sr)
        sr = re.sub(r'__main__|trac\.util\.tests(\.__init__)?', 'MODULE', sr)
        self.assertEqual("<MODULE.eh_ix object at 0xADDRESS "
                         "(repr() error: TypeError: unsupported operand "
                         "type(s) for +: 'int' and 'str')>", sr)


class SetuptoolsUtilsTestCase(unittest.TestCase):

    def setUp(self):
        self.dir = mkdtemp()
        sys.path.append(self.dir)

    def tearDown(self):
        sys.path.remove(self.dir)
        rmtree(self.dir)

    def test_get_module_path(self):
        self.assertEqual(util.get_module_path(trac),
                         util.get_module_path(util))

    def test_get_pkginfo_trac(self):
        pkginfo = util.get_pkginfo(trac)
        self.assertEqual(trac.__version__, pkginfo.get('version'))
        self.assertNotEqual({}, pkginfo)

    def test_get_pkginfo_non_toplevel(self):
        from trac import core
        import tracopt
        pkginfo = util.get_pkginfo(trac)
        self.assertEqual(pkginfo, util.get_pkginfo(util))
        self.assertEqual(pkginfo, util.get_pkginfo(core))
        self.assertEqual(pkginfo, util.get_pkginfo(tracopt))

    def test_get_pkginfo_genshi(self):
        try:
            import genshi
            import genshi.core
            dist = pkg_resources.get_distribution('Genshi')
        except:
            pass
        else:
            pkginfo = util.get_pkginfo(genshi)
            self.assertNotEqual({}, pkginfo)
            self.assertEqual(pkginfo, util.get_pkginfo(genshi.core))

    def test_get_pkginfo_babel(self):
        try:
            import babel
            import babel.core
            dist = pkg_resources.get_distribution('Babel')
        except:
            pass
        else:
            pkginfo = util.get_pkginfo(babel)
            self.assertNotEqual({}, pkginfo)
            self.assertEqual(pkginfo, util.get_pkginfo(babel.core))

    def test_get_pkginfo_pymysql(self):
        try:
            import pymysql
            dist = pkg_resources.get_distribution('pymysql')
            dist.get_metadata('top_level.txt')
        except:
            pass
        else:
            pkginfo = util.get_pkginfo(pymysql)
            self.assertNotEqual({}, pkginfo)
            self.assertEqual(pkginfo, util.get_pkginfo(pymysql.cursors))

    def test_get_pkginfo_psycopg2(self):
        # python-psycopg2 deb package doesn't provide SOURCES.txt and
        # top_level.txt
        try:
            import psycopg2
            import psycopg2.extensions
            dist = pkg_resources.get_distribution('psycopg2')
        except:
            pass
        else:
            pkginfo = util.get_pkginfo(psycopg2)
            self.assertNotEqual({}, pkginfo)
            self.assertEqual(pkginfo, util.get_pkginfo(psycopg2.extensions))

    def test_file_metadata(self):
        pkgname = 'TestModule_' + util.hex_entropy(16)
        modname = pkgname.lower()
        with open(os.path.join(self.dir, pkgname + '-0.1.egg-info'), 'w') as f:
            f.write('Metadata-Version: 1.1\n'
                    'Name: %(pkgname)s\n'
                    'Version: 0.1\n'
                    'Author: Joe\n'
                    'Author-email: joe@example.org\n'
                    'Maintainer: Jim\n'
                    'Maintainer-email: jim@example.org\n'
                    'Home-page: http://example.org/\n'
                    'Summary: summary.\n'
                    'Description: description.\n'
                    'Provides: %(modname)s\n'
                    'Provides: %(modname)s.foo\n'
                    % {'pkgname': pkgname, 'modname': modname})
        os.mkdir(os.path.join(self.dir, modname))
        for name in ('__init__.py', 'bar.py', 'foo.py'):
            with open(os.path.join(self.dir, modname, name), 'w') as f:
                f.write('# -*- coding: utf-8 -*-\n')

        mod = importlib.import_module(modname)
        mod.bar = importlib.import_module(modname + '.bar')
        mod.foo = importlib.import_module(modname + '.foo')
        pkginfo = util.get_pkginfo(mod)
        self.assertEqual('0.1', pkginfo['version'])
        self.assertEqual('Joe', pkginfo['author'])
        self.assertEqual('joe@example.org', pkginfo['author_email'])
        self.assertEqual('Jim', pkginfo['maintainer'])
        self.assertEqual('jim@example.org', pkginfo['maintainer_email'])
        self.assertEqual('http://example.org/', pkginfo['home_page'])
        self.assertEqual('summary.', pkginfo['summary'])
        self.assertEqual('description.', pkginfo['description'])
        self.assertEqual(pkginfo, util.get_pkginfo(mod.bar))
        self.assertEqual(pkginfo, util.get_pkginfo(mod.foo))

    def _write_module(self, version, url):
        modname = 'TestModule_' + util.hex_entropy(16)
        modpath = os.path.join(self.dir, modname + '.py')
        with open(modpath, 'w') as f:
            f.write(textwrap.dedent("""\
                # -*- coding: utf-8 -*-
                from trac.core import Component

                version = '%s'
                author = 'Joe'
                author_email = 'joe@example.org'
                maintainer = 'Jim'
                maintainer_email = 'jim@example.org'
                home_page = '%s'
                license = 'BSD 3-Clause'
                summary = 'summary.'
                trac = 'http://my.trac.com'

                class TestModule(Component):
                    pass
                """) % (version, url))
        return modname

    def test_get_module_metadata(self):
        version = '0.1'
        home_page = 'http://example.org'
        modname = self._write_module(version, home_page)

        mod = importlib.import_module(modname)
        info = util.get_module_metadata(mod)

        self.assertEqual(version, info['version'])
        self.assertEqual('Joe', info['author'])
        self.assertEqual('joe@example.org', info['author_email'])
        self.assertEqual('Jim', info['maintainer'])
        self.assertEqual('jim@example.org', info['maintainer_email'])
        self.assertEqual(home_page, info['home_page'])
        self.assertEqual('summary.', info['summary'])
        self.assertEqual('BSD 3-Clause', info['license'])
        self.assertEqual('http://my.trac.com', info['trac'])

    def test_get_module_metadata_keyword_expansion(self):
        version = '10'
        url = 'http://example.org'
        modname = self._write_module('$Rev: %s $' % version,
                                     '$URL: %s $' % url)

        mod = importlib.import_module(modname)
        info = util.get_module_metadata(mod)

        self.assertEqual('r%s' % version, info['version'])
        self.assertEqual(url, info['home_page'])


class LazyClass(object):
    @util.lazy
    def f(self):
        return object()


class LazyTestCase(unittest.TestCase):

    def setUp(self):
        self.obj = LazyClass()

    def test_lazy_get(self):
        f = self.obj.f
        self.assertTrue(self.obj.f is f)

    def test_lazy_set(self):
        self.obj.f = 2
        self.assertEqual(2, self.obj.f)

    def test_lazy_del(self):
        f = self.obj.f
        del self.obj.f
        self.assertFalse(self.obj.f is f)


class FileTestCase(unittest.TestCase):

    def setUp(self):
        self.dir = mkdtemp()
        self.filename = os.path.join(self.dir, 'trac-tempfile')
        self.data = 'Lorem\ripsum\ndolor\r\nsit\namet,\rconsectetur\r\n'

    def tearDown(self):
        rmtree(self.dir)

    def test_create_and_read_file(self):
        util.create_file(self.filename, self.data, 'wb')
        with open(self.filename, 'rb') as f:
            self.assertEqual(self.data, f.read())
        self.assertEqual(self.data, util.read_file(self.filename, 'rb'))

    def test_touch_file(self):
        util.create_file(self.filename, self.data, 'wb')
        util.touch_file(self.filename)
        with open(self.filename, 'rb') as f:
            self.assertEqual(self.data, f.read())

    def test_missing(self):
        util.touch_file(self.filename)
        self.assertTrue(os.path.isfile(self.filename))
        self.assertEqual(0, os.path.getsize(self.filename))

class UtilitiesTestCase(unittest.TestCase):

    def test_as_int(self):
        self.assertEqual(1, util.as_int('1'))
        self.assertEqual(1, util.as_int('1', None))
        self.assertIsNone(util.as_int('A', None))
        self.assertEqual(2, util.as_int('A', 2))
        self.assertEqual(2, util.as_int('1', None, min=2))
        self.assertEqual(0, util.as_int('1', None, max=0))

    def test_as_float(self):
        self.assertEqual(1.1, util.as_float('1.1'))
        self.assertEqual(1.1, util.as_float('1.1', None))
        self.assertEqual(1, util.as_float('1', None))
        self.assertIsNone(util.as_float('A', None))
        self.assertEqual(2.2, util.as_float('A', 2.2))
        self.assertEqual(2.2, util.as_float('1.1', None, min=2.2))
        self.assertEqual(0.1, util.as_float('1.1', None, max=0.1))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(AtomicFileTestCase))
    suite.addTest(unittest.makeSuite(PathTestCase))
    suite.addTest(unittest.makeSuite(RandomTestCase))
    suite.addTest(unittest.makeSuite(ContentDispositionTestCase))
    suite.addTest(unittest.makeSuite(SafeReprTestCase))
    suite.addTest(unittest.makeSuite(SetuptoolsUtilsTestCase))
    suite.addTest(unittest.makeSuite(LazyTestCase))
    suite.addTest(unittest.makeSuite(FileTestCase))
    suite.addTest(unittest.makeSuite(UtilitiesTestCase))
    suite.addTest(concurrency.test_suite())
    suite.addTest(datefmt.test_suite())
    suite.addTest(presentation.test_suite())
    suite.addTest(doctest.DocTestSuite(util))
    suite.addTest(text.test_suite())
    suite.addTest(translation.test_suite())
    suite.addTest(html.test_suite())
    suite.addTest(doctest.DocTestSuite(util.html))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2023 Edgewall Software
# Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.
#
# Author: Christopher Lenz <cmlenz@gmx.de>
from datetime import datetime
import io
import locale
import os.path
import textwrap
import unittest

try:
    from svn import core, repos
    has_svn = True
except ImportError:
    has_svn = False

from trac.test import EnvironmentStub, MockRequest, TestSetup, mkdtemp
from trac.core import TracError
from trac.mimeview.api import RenderingContext
from trac.resource import Resource, resource_exists
from trac.util.concurrency import get_thread_id
from trac.util.datefmt import utc
from trac.util.html import Element, Fragment, Markup
from trac.versioncontrol.api import DbRepositoryProvider, Changeset, \
                                    InvalidRepository, Node, \
                                    NoSuchChangeset, RepositoryManager
from trac.web.chrome import web_context
from tracopt.versioncontrol.svn import svn_fs, svn_prop


REPOS_PATH = None
REPOS_NAME = 'repo'
URL = 'svn://test'

HEAD = 31
TETE = 26

NATIVE_EOL = '\r\n' if os.name == 'nt' else '\n'


def _create_context(env):
    req = MockRequest(env)
    return web_context(req)


def setlocale(*locales):
    """Decorator to call test method with each locale.
    """
    if os.name != 'nt':
        locales_map = {'de': 'de_DE.UTF8', 'fr': 'fr_FR.UTF8',
                       'pl': 'pl_PL.UTF8', 'ja': 'ja_JP.UTF8',
                       'zh_CN': 'zh_CN.UTF8'}
    else:
        locales_map = {'de': 'German_Germany', 'fr': 'French_France',
                       'pl': 'Polish_Poland', 'ja': 'Japanese_Japan',
                       'zh_CN': "Chinese_People's Republic of China"}
    locales_map['C'] = 'C'

    def getlocale():
        rv = locale.getlocale(locale.LC_ALL)
        return rv if rv[0] else 'C'

    def setlocale(locale_id):
        locale_id = locales_map[locale_id]
        try:
            locale.setlocale(locale.LC_ALL, locale_id)
            return True
        except locale.Error:
            return False

    def wrap(fn, locales):
        def wrapped(*args, **kwargs):
            saved_locale = getlocale()
            try:
                for locale_id in locales:
                    if setlocale(locale_id):
                        fn(*args, **kwargs)
            finally:
                locale.setlocale(locale.LC_ALL, saved_locale)
        return wrapped

    if len(locales) == 1 and hasattr(locales[0], '__call__'):
        return wrap(locales[0], sorted(locales_map))
    else:
        def decorator(fn):
            return wrap(fn, locales)
        return decorator


class SubversionRepositoryTestSetup(TestSetup):

    def setUp(self):
        dumpfile = open(os.path.join(os.path.split(__file__)[0],
                                     'svnrepos.dump'), 'rb')

        svn_fs._import_svn()
        core.apr_initialize()
        pool = core.svn_pool_create(None)
        dumpstream = None
        try:
            r = repos.svn_repos_create(REPOS_PATH, '', '', None, None, pool)
            if hasattr(repos, 'svn_repos_load_fs2'):
                repos.svn_repos_load_fs2(r, dumpfile, io.BytesIO(),
                                         repos.svn_repos_load_uuid_default,
                                         '', 0, 0, None, pool)
            else:
                dumpstream = core.svn_stream_from_aprfile(dumpfile, pool)
                repos.svn_repos_load_fs(r, dumpstream, None,
                                        repos.svn_repos_load_uuid_default, '',
                                        None, None, pool)
        finally:
            dumpfile.close()
            if dumpstream:
                core.svn_stream_close(dumpstream)
            core.svn_pool_destroy(pool)
            core.apr_terminate()

    def tearDown(self):
        repos.svn_repos_delete(REPOS_PATH)


# -- Re-usable test mixins

class NormalTests(object):

    def _row_col(self, elt, row, col):
        row_idx = 1
        if not isinstance(elt, Fragment):
            return
        for tr in elt.children:
            if isinstance(tr, Element) and tr.tag == 'tr':
                if row_idx == row:
                    col_idx = 1
                    for td in tr.children:
                        if isinstance(td, Element) and td.tag == 'td':
                            if col_idx == col:
                                return unicode(td)
                            col_idx += 1
                row_idx += 1
            else:
                td = self._row_col(tr, row, col) # recurse
                if td:
                    return td

    def test_invalid_path_raises(self):
        self.assertRaises(InvalidRepository, svn_fs.SubversionRepository,
                          '/the/invalid/path', [], self.env.log)

    def test_resource_exists(self):
        resource = Resource('repository', REPOS_NAME)
        self.assertTrue(resource_exists(self.env, resource))
        self.assertFalse(resource_exists(self.env, resource(id='xxx')))
        node = resource.child('source', u'tête')
        self.assertTrue(resource_exists(self.env, node))
        self.assertFalse(resource_exists(self.env, node(id='xxx')))
        cset = resource.child('changeset', HEAD)
        self.assertTrue(resource_exists(self.env, cset))
        self.assertFalse(resource_exists(self.env, cset(id=123456)))

    def test_repos_normalize_path(self):
        self.assertEqual('/', self.repos.normalize_path('/'))
        self.assertEqual('/', self.repos.normalize_path(''))
        self.assertEqual('/', self.repos.normalize_path(None))
        self.assertEqual(u'tête', self.repos.normalize_path(u'tête'))
        self.assertEqual(u'tête', self.repos.normalize_path(u'/tête'))
        self.assertEqual(u'tête', self.repos.normalize_path(u'tête/'))
        self.assertEqual(u'tête', self.repos.normalize_path(u'/tête/'))

    def test_repos_normalize_rev(self):
        self.assertEqual(HEAD, self.repos.normalize_rev('latest'))
        self.assertEqual(HEAD, self.repos.normalize_rev('head'))
        self.assertEqual(HEAD, self.repos.normalize_rev(''))
        self.assertRaises(NoSuchChangeset,
                          self.repos.normalize_rev, 'something else')
        self.assertEqual(HEAD, self.repos.normalize_rev(None))
        self.assertEqual(11, self.repos.normalize_rev('11'))
        self.assertEqual(11, self.repos.normalize_rev(11))
        self.assertRaises(NoSuchChangeset, self.repos.normalize_rev, -1)
        self.assertRaises(NoSuchChangeset, self.repos.normalize_rev, -42)

    def test_repos_display_rev(self):
        self.assertEqual(str(HEAD), self.repos.display_rev('latest'))
        self.assertEqual(str(HEAD), self.repos.display_rev('head'))
        self.assertEqual(str(HEAD), self.repos.display_rev(''))
        self.assertRaises(NoSuchChangeset,
                          self.repos.display_rev, 'something else')
        self.assertEqual(str(HEAD), self.repos.display_rev(None))
        self.assertEqual('11', self.repos.display_rev('11'))
        self.assertEqual('11', self.repos.display_rev(11))

    def test_repos_short_rev(self):
        self.assertEqual(str(HEAD), self.repos.short_rev('latest'))
        self.assertEqual(str(HEAD), self.repos.short_rev('head'))
        self.assertEqual(str(HEAD), self.repos.short_rev(''))
        self.assertRaises(NoSuchChangeset,
                          self.repos.short_rev, 'something else')
        self.assertEqual(str(HEAD), self.repos.short_rev(None))
        self.assertEqual('11', self.repos.short_rev('11'))
        self.assertEqual('11', self.repos.short_rev(11))

    def test_rev_navigation(self):
        self.assertEqual(1, self.repos.oldest_rev)
        self.assertIsNone(self.repos.previous_rev(0))
        self.assertIsNone(self.repos.previous_rev(1))
        self.assertEqual(HEAD, self.repos.youngest_rev)
        self.assertEqual(6, self.repos.next_rev(5))
        self.assertEqual(7, self.repos.next_rev(6))
        # ...
        self.assertIsNone(self.repos.next_rev(HEAD))
        self.assertRaises(NoSuchChangeset, self.repos.normalize_rev, HEAD + 1)

    def test_rev_path_navigation(self):
        self.assertEqual(1, self.repos.oldest_rev)
        self.assertIsNone(self.repos.previous_rev(0, u'tête'))
        self.assertIsNone(self.repos.previous_rev(1, u'tête'))
        self.assertEqual(HEAD, self.repos.youngest_rev)
        self.assertEqual(6, self.repos.next_rev(5, u'tête'))
        self.assertEqual(13, self.repos.next_rev(6, u'tête'))
        # ...
        self.assertIsNone(self.repos.next_rev(HEAD, u'tête'))
        # test accentuated characters
        self.assertIsNone(self.repos.previous_rev(17, u'tête/R\xe9sum\xe9.txt'))
        self.assertEqual(17, self.repos.next_rev(16, u'tête/R\xe9sum\xe9.txt'))

    def test_has_node(self):
        self.assertFalse(self.repos.has_node(u'/tête/dir1', 3))
        self.assertTrue(self.repos.has_node(u'/tête/dir1', 4))
        self.assertTrue(self.repos.has_node(u'/tête/dir1'))

    def test_get_node(self):
        node = self.repos.get_node(u'/')
        self.assertEqual(u'', node.name)
        self.assertEqual(u'/', node.path)
        self.assertEqual(Node.DIRECTORY, node.kind)
        self.assertEqual(HEAD, node.rev)
        self.assertEqual(HEAD, node.created_rev)
        self.assertEqual(datetime(2017, 1, 9, 6, 12, 33, 247657, utc),
                         node.last_modified)
        self.assertRaises(NoSuchChangeset, self.repos.get_node, u'/', -1)
        node = self.repos.get_node(u'/tête')
        self.assertEqual(u'tête', node.name)
        self.assertEqual(u'/tête', node.path)
        self.assertEqual(Node.DIRECTORY, node.kind)
        self.assertEqual(HEAD, node.rev)
        self.assertEqual(TETE, node.created_rev)
        self.assertEqual(datetime(2013, 4, 28, 5, 36, 6, 29637, utc),
                         node.last_modified)
        node = self.repos.get_node(u'/tête/README.txt')
        self.assertEqual('README.txt', node.name)
        self.assertEqual(u'/tête/README.txt', node.path)
        self.assertEqual(Node.FILE, node.kind)
        self.assertEqual(HEAD, node.rev)
        self.assertEqual(3, node.created_rev)
        self.assertEqual(datetime(2005, 4, 1, 13, 24, 58, 234643, utc),
                         node.last_modified)

    def test_get_node_specific_rev(self):
        node = self.repos.get_node(u'/tête', 1)
        self.assertEqual(u'tête', node.name)
        self.assertEqual(u'/tête', node.path)
        self.assertEqual(Node.DIRECTORY, node.kind)
        self.assertEqual(1, node.rev)
        self.assertEqual(datetime(2005, 4, 1, 10, 0, 52, 353248, utc),
                         node.last_modified)
        node = self.repos.get_node(u'/tête/README.txt', 2)
        self.assertEqual('README.txt', node.name)
        self.assertEqual(u'/tête/README.txt', node.path)
        self.assertEqual(Node.FILE, node.kind)
        self.assertEqual(2, node.rev)
        self.assertEqual(datetime(2005, 4, 1, 13, 12, 18, 216267, utc),
                         node.last_modified)

    def test_get_dir_entries(self):
        node = self.repos.get_node(u'/tête')
        entries = node.get_entries()
        self.assertEqual('dir1', next(entries).name)
        self.assertEqual('mpp_proc', next(entries).name)
        self.assertEqual('v2', next(entries).name)
        self.assertEqual('README3.txt', next(entries).name)
        self.assertEqual(u'R\xe9sum\xe9.txt', next(entries).name)
        self.assertEqual('README.txt', next(entries).name)
        self.assertRaises(StopIteration, next, entries)

    def test_get_file_entries(self):
        node = self.repos.get_node(u'/tête/README.txt')
        entries = node.get_entries()
        self.assertRaises(StopIteration, next, entries)

    def test_get_dir_content(self):
        node = self.repos.get_node(u'/tête')
        self.assertIsNone(node.content_length)
        self.assertIsNone(node.content_type)
        self.assertIsNone(node.get_content())

    def test_get_file_content(self):
        node = self.repos.get_node(u'/tête/README.txt')
        self.assertEqual(8, node.content_length)
        self.assertEqual('text/plain', node.content_type)
        self.assertEqual('A test.\n', node.get_content().read())

    def test_get_dir_properties(self):
        f = self.repos.get_node(u'/tête')
        props = f.get_properties()
        self.assertEqual(1, len(props))

    def test_get_file_properties(self):
        f = self.repos.get_node(u'/tête/README.txt')
        props = f.get_properties()
        self.assertEqual('native', props['svn:eol-style'])
        self.assertEqual('text/plain', props['svn:mime-type'])

    def test_get_file_content_without_native_eol_style(self):
        f = self.repos.get_node(u'/tête/README.txt', 2)
        props = f.get_properties()
        self.assertIsNone(props.get('svn:eol-style'))
        self.assertEqual('A text.\n', f.get_content().read())
        self.assertEqual('A text.\n', f.get_processed_content().read())

    def test_get_file_content_with_native_eol_style(self):
        f = self.repos.get_node(u'/tête/README.txt', 3)
        props = f.get_properties()
        self.assertEqual('native', props.get('svn:eol-style'))

        self.repos.params['eol_style'] = 'native'
        self.assertEqual('A test.\n', f.get_content().read())
        self.assertEqual('A test.' + NATIVE_EOL,
                         f.get_processed_content().read())

        self.repos.params['eol_style'] = 'LF'
        self.assertEqual('A test.\n', f.get_content().read())
        self.assertEqual('A test.\n', f.get_processed_content().read())

        self.repos.params['eol_style'] = 'CRLF'
        self.assertEqual('A test.\n', f.get_content().read())
        self.assertEqual('A test.\r\n', f.get_processed_content().read())

        self.repos.params['eol_style'] = 'CR'
        self.assertEqual('A test.\n', f.get_content().read())
        self.assertEqual('A test.\r', f.get_processed_content().read())
        # check that the hint is stronger than the repos default
        self.assertEqual('A test.\r\n',
                         f.get_processed_content(eol_hint='CRLF').read())

    def test_get_file_content_with_native_eol_style_and_no_keywords_28(self):
        f = self.repos.get_node(u'/branches/v4/README.txt', 28)
        props = f.get_properties()
        self.assertEqual('native', props.get('svn:eol-style'))
        self.assertIsNone(props.get('svn:keywords'))

        self.assertEqual(
            'A test.\n' +
            '# $Rev$ is not substituted with no svn:keywords.\n',
            f.get_content().read())
        self.assertEqual(
            'A test.\r\n' +
            '# $Rev$ is not substituted with no svn:keywords.\r\n',
            f.get_processed_content(eol_hint='CRLF').read())

    def test_get_file_content_with_keyword_substitution_23(self):
        f = self.repos.get_node(u'/tête/Résumé.txt', 23)
        props = f.get_properties()
        self.assertEqual('Revision Author URL', props['svn:keywords'])
        self.assertEqual('''\
# Simple test for svn:keywords property substitution (#717)
# $Rev: 23 $:     Revision of last commit
# $Author: cboos $:  Author of last commit
# $Date$:    Date of last commit (not substituted)

Now with fixed width fields:
# $URL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.txt    $ the configured URL
# $HeadURL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.tx#$ same
# $URL:: svn://test/t%C#$ same, but truncated

En r\xe9sum\xe9 ... \xe7a marche.
'''.splitlines(), f.get_processed_content().read().splitlines())
    # Note: "En résumé ... ça marche." in the content is really encoded in
    #       latin1 in the file, and our substitutions are UTF-8 encoded...
    #       This is expected.

    def test_get_file_content_with_keyword_substitution_24(self):
        f = self.repos.get_node(u'/tête/Résumé.txt', 24)
        props = f.get_properties()
        self.assertEqual('Revision Author URL Id', props['svn:keywords'])
        self.assertEqual('''\
# Simple test for svn:keywords property substitution (#717)
# $Rev: 24 $:     Revision of last commit
# $Author: cboos $:  Author of last commit
# $Date$:    Date of last commit (now substituted)
# $Id: Résumé.txt 24 2013-04-27 14:38:50Z cboos $:      Combination

Now with fixed width fields:
# $URL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.txt    $ the configured URL
# $HeadURL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.tx#$ same
# $URL:: svn://test/t%C#$ same, but truncated
# $Header::                                           $ combination with URL

En r\xe9sum\xe9 ... \xe7a marche.
'''.splitlines(), f.get_processed_content().read().splitlines())

    @setlocale
    def test_get_file_content_with_keyword_substitution_25(self):
        f = self.repos.get_node(u'/tête/Résumé.txt', 25)
        props = f.get_properties()
        self.assertEqual('Revision Author URL Date Id Header',
                         props['svn:keywords'])
        self.assertEqual('''\
# Simple test for svn:keywords property substitution (#717)
# $Rev: 25 $:     Revision of last commit
# $Author: cboos $:  Author of last commit
# $Date: 2013-04-27 14:43:15 +0000 (Sat, 27 Apr 2013) $:    Date of last commit (now really substituted)
# $Id: Résumé.txt 25 2013-04-27 14:43:15Z cboos $:      Combination

Now with fixed width fields:
# $URL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.txt    $ the configured URL
# $HeadURL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.tx#$ same
# $URL:: svn://test/t%C#$ same, but truncated
# $Header:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.txt#$ combination with URL

En r\xe9sum\xe9 ... \xe7a marche.
'''.splitlines(), f.get_processed_content().read().splitlines())

    @setlocale
    def test_get_file_content_with_keyword_substitution_30(self):
        self.maxDiff = None
        f = self.repos.get_node(u'/branches/v4/Résumé.txt', 30)
        expected = [
            '# Simple test for svn:keywords property substitution (#717)',
            '# $Rev: 30 $:     Revision of last commit',
            '# $Author: jomae $:  Author of last commit',
            '# $Date: 2015-06-15 14:09:13 +0000 (Mon, 15 Jun 2015) $:    '
                'Date of last commit (now really substituted)',
            '# $Id: Résumé.txt 30 2015-06-15 14:09:13Z jomae $:      '
                'Combination',
            '',
            'Now with fixed width fields:',
            '# $URL:: svn://test/branches/v4/R%C3%A9sum%C3%A9.txt  $ '
                'the configured URL',
            '# $HeadURL:: svn://test/branches/v4/R%C3%A9sum%C3%A9.#$ same',
            '# $URL:: svn://test/bra#$ same, but truncated',
            '# $Header:: svn://test/branches/v4/R%C3%A9sum%C3%A9.t#$ '
                'combination with URL',
            '',
            'Overlapped keywords:',
            '# $Xxx$Rev: 30 $Xxx$',
            '# $Rev: 30 $Xxx$Rev: 30 $',
            '# $Rev: 30 $Rev$Rev: 30 $',
            '',
            'Custom keyword definitions (#11364)',
            '# $_Author: jomae $:',
            '# $_Basename: R\xc3\xa9sum\xc3\xa9.txt $:',
            '# $_ShortDate: 2015-06-15 14:09:13Z $:',
            '# $_LongDate: 2015-06-15 14:09:13 +0000 (Mon, 15 Jun 2015) $:',
            '# $_Path: branches/v4/R\xc3\xa9sum\xc3\xa9.txt $:',
            '# $_Rev: 30 $:',
            '# $_RootURL: svn://test $:',
            '# $_URL: svn://test/branches/v4/R%C3%A9sum%C3%A9.txt $:',
            '# $_Header: branches/v4/R\xc3\xa9sum\xc3\xa9.txt 30 '
                '2015-06-15 14:09:13Z jomae $:',
            '# $_Id: R\xc3\xa9sum\xc3\xa9.txt 30 '
                '2015-06-15 14:09:13Z jomae $:',
            '# $_Header2: branches/v4/R\xc3\xa9sum\xc3\xa9.txt 30 '
                '2015-06-15 14:09:13Z jomae $:',
            '# $_Id2: R\xc3\xa9sum\xc3\xa9.txt 30 '
                '2015-06-15 14:09:13Z jomae $:',
            '# $_t\xc3\xa9t\xc3\xa9: jomae $:',
            '# $42: jomae $:',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '123456789012345678901234567890123456789: j $:',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '1234567890123456789012345678901234567890:  $:',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901$:',
            '# $_TooLong: branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'branches/v4/R\xc3\xa9sum\xc3\xa9.txt'
                'br $:',
            '',
            'Custom keyword definitions with fixed width',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '1234567890123456789012345678901234:: jomae $',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '1234567890123456789012345678901234::        $',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678:: j#$',
            '# $123456789012345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678901234567890'
                '12345678901234567890123456789012345678::    $',
        ]
        self.assertEqual(expected,
                         f.get_processed_content().read().splitlines())

    @setlocale
    def test_get_file_content_with_keyword_substitution_31(self):
        """$Author$ is is the last user to change this file in the repository.
        Regression test for #12655.
        """
        self.assertEqual('john', self.repos.get_changeset(31).author)
        self.assertEqual('jomae', self.repos.get_changeset(26).author)
        f = self.repos.get_node(u'/tête/Résumé.txt', 31)
        self.assertEqual(31, f.rev)
        self.assertEqual(26, f.created_rev)
        props = f.get_properties()
        self.assertEqual('Revision Author URL Date Id Header',
                         props['svn:keywords'])
        self.assertEqual('''\
# Simple test for svn:keywords property substitution (#717)
# $Rev: 26 $:     Revision of last commit
# $Author: jomae $:  Author of last commit
# $Date: 2013-04-28 05:36:06 +0000 (Sun, 28 Apr 2013) $:    Date of last commit (now really substituted)
# $Id: Résumé.txt 26 2013-04-28 05:36:06Z jomae $:      Combination

Now with fixed width fields:
# $URL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.txt    $ the configured URL
# $HeadURL:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.tx#$ same
# $URL:: svn://test/t%C#$ same, but truncated
# $Header:: svn://test/t%C3%AAte/R%C3%A9sum%C3%A9.txt#$ combination with URL

Overlapped keywords:
# $Xxx$Rev: 26 $Xxx$
# $Rev: 26 $Xxx$Rev: 26 $
# $Rev: 26 $Rev$Rev: 26 $

En r\xe9sum\xe9 ... \xe7a marche.
'''.splitlines(), f.get_processed_content().read().splitlines())

    def test_created_path_rev(self):
        node = self.repos.get_node(u'/tête/README3.txt', 15)
        self.assertEqual(15, node.rev)
        self.assertEqual(u'/tête/README3.txt', node.path)
        self.assertEqual(14, node.created_rev)
        self.assertEqual(u'tête/README3.txt', node.created_path)

    def test_created_path_rev_parent_copy(self):
        node = self.repos.get_node('/tags/v1/README.txt', 15)
        self.assertEqual(15, node.rev)
        self.assertEqual('/tags/v1/README.txt', node.path)
        self.assertEqual(3, node.created_rev)
        self.assertEqual(u'tête/README.txt', node.created_path)

    def test_get_annotations(self):
        # svn_client_blame2() requires a canonical uri since Subversion 1.7.
        # If the uri is not canonical, assertion raises (#11167).
        node = self.repos.get_node(u'/tête/R\xe9sum\xe9.txt', 25)
        self.assertEqual([23, 23, 23, 25, 24, 23, 23, 23, 23, 23, 24, 23, 20],
                         node.get_annotations())

    def test_get_annotations_lower_drive_letter(self):
        # If the drive letter in the uri is lower case on Windows, a
        # SubversionException raises (#10514).
        drive, tail = os.path.splitdrive(REPOS_PATH)
        repos_path = drive.lower() + tail
        DbRepositoryProvider(self.env).add_repository('lowercase', repos_path,
                                                      'direct-svnfs')
        repos = RepositoryManager(self.env).get_repository('lowercase')
        node = repos.get_node(u'/tête/R\xe9sum\xe9.txt', 25)
        self.assertEqual([23, 23, 23, 25, 24, 23, 23, 23, 23, 23, 24, 23, 20],
                         node.get_annotations())

    if os.name != 'nt':
        del test_get_annotations_lower_drive_letter

    def test_get_annotations_with_urlencoded_percent_sign(self):
        node = self.repos.get_node(u'/branches/t10386/READ%25ME.txt')
        self.assertEqual([14], node.get_annotations())

    def test_get_path_url(self):
        self.assertEqual('svn://test', self.repos.get_path_url('', 42))
        self.assertEqual('svn://test', self.repos.get_path_url('/', 42))
        self.assertEqual('svn://test/path/to/file.txt',
                         self.repos.get_path_url('path/to/file.txt', 42))
        self.assertEqual('svn://test/path/to/file.txt',
                         self.repos.get_path_url('/path/to/file.txt', 42))
        self.assertEqual('svn://test/trunk%25/Resume%25.txt',
                         self.repos.get_path_url('trunk%/Resume%.txt', 42))
        self.assertEqual('svn://test/trunk%23/Resume%23.txt',
                         self.repos.get_path_url('trunk#/Resume#.txt', 42))
        self.assertEqual('svn://test/trunk%3F/Resume%3F.txt',
                         self.repos.get_path_url('trunk?/Resume?.txt', 42))
        self.assertEqual('svn://test/trunk%40/Resume.txt%4042',
                         self.repos.get_path_url('trunk@/Resume.txt@42', 42))
        self.assertEqual('svn://test/tr%C3%BCnk/R%C3%A9sum%C3%A9.txt',
                         self.repos.get_path_url(u'trünk/Résumé.txt', 42))

    # Revision Log / node history

    def test_get_node_history(self):
        node = self.repos.get_node(u'/tête/README3.txt')
        history = node.get_history()
        self.assertEqual((u'tête/README3.txt', 14, 'copy'), next(history))
        self.assertEqual((u'tête/README2.txt', 6, 'copy'), next(history))
        self.assertEqual((u'tête/README.txt', 3, 'edit'), next(history))
        self.assertEqual((u'tête/README.txt', 2, 'add'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_node_history_limit(self):
        node = self.repos.get_node(u'/tête/README3.txt')
        history = node.get_history(2)
        self.assertEqual((u'tête/README3.txt', 14, 'copy'), next(history))
        self.assertEqual((u'tête/README2.txt', 6, 'copy'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_node_history_follow_copy(self):
        node = self.repos.get_node('/tags/v1/README.txt')
        history = node.get_history()
        self.assertEqual(('tags/v1/README.txt', 7, 'copy'), next(history))
        self.assertEqual((u'tête/README.txt', 3, 'edit'), next(history))
        self.assertEqual((u'tête/README.txt', 2, 'add'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_copy_ancestry(self):
        node = self.repos.get_node('/tags/v1/README.txt')
        ancestry = node.get_copy_ancestry()
        self.assertEqual([(u'tête/README.txt', 6)], ancestry)
        for path, rev in ancestry:
            self.repos.get_node(path, rev)  # shouldn't raise NoSuchNode

        node = self.repos.get_node(u'/tête/README3.txt')
        ancestry = node.get_copy_ancestry()
        self.assertEqual([(u'tête/README2.txt', 13),
                          (u'tête/README.txt', 3)], ancestry)
        for path, rev in ancestry:
            self.repos.get_node(path, rev)  # shouldn't raise NoSuchNode

        node = self.repos.get_node('/branches/v1x')
        ancestry = node.get_copy_ancestry()
        self.assertEqual([(u'tags/v1.1', 11),
                          (u'branches/v1x', 9),
                          (u'tags/v1', 7),
                          (u'tête', 6)], ancestry)
        for path, rev in ancestry:
            self.repos.get_node(path, rev)  # shouldn't raise NoSuchNode

    def test_get_copy_ancestry_for_move(self):
        node = self.repos.get_node(u'/tête/dir1/dir2', 5)
        ancestry = node.get_copy_ancestry()
        self.assertEqual([(u'tête/dir2', 4)], ancestry)
        for path, rev in ancestry:
            self.repos.get_node(path, rev)  # shouldn't raise NoSuchNode

    def test_get_branch_origin(self):
        node = self.repos.get_node('/tags/v1/README.txt')
        self.assertEqual(7, node.get_branch_origin())
        node = self.repos.get_node(u'/tête/README3.txt')
        self.assertEqual(14, node.get_branch_origin())
        node = self.repos.get_node('/branches/v1x')
        self.assertEqual(12, node.get_branch_origin())
        node = self.repos.get_node(u'/tête/dir1/dir2', 5)
        self.assertEqual(5, node.get_branch_origin())

    # Revision Log / path history

    def test_get_path_history(self):
        history = self.repos.get_path_history(u'/tête/README2.txt', None)
        self.assertEqual((u'tête/README2.txt', 14, 'delete'), next(history))
        self.assertEqual((u'tête/README2.txt', 6, 'copy'), next(history))
        self.assertEqual((u'tête/README.txt', 3, 'unknown'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_path_history_copied_file(self):
        history = self.repos.get_path_history('/tags/v1/README.txt', None)
        self.assertEqual(('tags/v1/README.txt', 7, 'copy'), next(history))
        self.assertEqual((u'tête/README.txt', 3, 'unknown'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_path_history_copied_dir(self):
        history = self.repos.get_path_history('/branches/v1x', None)
        self.assertEqual(('branches/v1x', 12, 'copy'), next(history))
        self.assertEqual(('tags/v1.1', 10, 'unknown'), next(history))
        self.assertEqual(('branches/v1x', 11, 'delete'), next(history))
        self.assertEqual(('branches/v1x', 9, 'edit'), next(history))
        self.assertEqual(('branches/v1x', 8, 'copy'), next(history))
        self.assertEqual(('tags/v1', 7, 'unknown'), next(history))
        self.assertRaises(StopIteration, next, history)

    # Diffs

    def _cmp_diff(self, expected, got):
        if expected[0]:
            old = self.repos.get_node(*expected[0])
            self.assertEqual((old.path, old.rev), (got[0].path, got[0].rev))
        if expected[1]:
            new = self.repos.get_node(*expected[1])
            self.assertEqual((new.path, new.rev), (got[1].path, got[1].rev))
        self.assertEqual(expected[2], (got[2], got[3]))

    def test_diff_file_different_revs(self):
        diffs = self.repos.get_changes(u'tête/README.txt', 2,
                                       u'tête/README.txt', 3)
        self._cmp_diff(((u'tête/README.txt', 2),
                        (u'tête/README.txt', 3),
                        (Node.FILE, Changeset.EDIT)), next(diffs))
        self.assertRaises(StopIteration, next, diffs)

    def test_diff_file_different_files(self):
        diffs = self.repos.get_changes('branches/v1x/README.txt', 12,
                                       'branches/v1x/README2.txt', 12)
        self._cmp_diff((('branches/v1x/README.txt', 12),
                        ('branches/v1x/README2.txt', 12),
                        (Node.FILE, Changeset.EDIT)), next(diffs))
        self.assertRaises(StopIteration, next, diffs)

    def test_diff_file_no_change(self):
        diffs = self.repos.get_changes(u'tête/README.txt', 7,
                                       'tags/v1/README.txt', 7)
        self.assertRaises(StopIteration, next, diffs)

    def test_diff_dir_different_revs(self):
        diffs = self.repos.get_changes(u'tête', 4, u'tête', 8)
        self._cmp_diff((None, (u'tête/README2.txt', 8),
                        (Node.FILE, Changeset.ADD)), next(diffs))
        self._cmp_diff((None, (u'tête/dir1/dir2', 8),
                        (Node.DIRECTORY, Changeset.ADD)), next(diffs))
        self._cmp_diff((None, (u'tête/dir1/dir3', 8),
                        (Node.DIRECTORY, Changeset.ADD)), next(diffs))
        self._cmp_diff(((u'tête/dir2', 4), None,
                        (Node.DIRECTORY, Changeset.DELETE)), next(diffs))
        self._cmp_diff(((u'tête/dir3', 4), None,
                        (Node.DIRECTORY, Changeset.DELETE)), next(diffs))
        self.assertRaises(StopIteration, next, diffs)

    def test_diff_dir_different_dirs(self):
        diffs = self.repos.get_changes(u'tête', 1, 'branches/v1x', 12)
        self._cmp_diff((None, ('branches/v1x/README.txt', 12),
                        (Node.FILE, Changeset.ADD)), next(diffs))
        self._cmp_diff((None, ('branches/v1x/README2.txt', 12),
                        (Node.FILE, Changeset.ADD)), next(diffs))
        self._cmp_diff((None, ('branches/v1x/dir1', 12),
                        (Node.DIRECTORY, Changeset.ADD)), next(diffs))
        self._cmp_diff((None, ('branches/v1x/dir1/dir2', 12),
                        (Node.DIRECTORY, Changeset.ADD)), next(diffs))
        self._cmp_diff((None, ('branches/v1x/dir1/dir3', 12),
                        (Node.DIRECTORY, Changeset.ADD)), next(diffs))
        self.assertRaises(StopIteration, next, diffs)

    def test_diff_dir_no_change(self):
        diffs = self.repos.get_changes(u'tête', 7, 'tags/v1', 7)
        self.assertRaises(StopIteration, next, diffs)

    # Changesets

    def test_changeset_repos_creation(self):
        chgset = self.repos.get_changeset(0)
        self.assertEqual(0, chgset.rev)
        self.assertEqual('', chgset.message)
        self.assertEqual('', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 9, 57, 41, 312767, utc),
                         chgset.date)
        self.assertRaises(StopIteration, next, chgset.get_changes())

    def test_changeset_added_dirs(self):
        chgset = self.repos.get_changeset(1)
        self.assertEqual(1, chgset.rev)
        self.assertEqual('Initial directory layout.', chgset.message)
        self.assertEqual('john', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 10, 0, 52, 353248, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual(('branches', Node.DIRECTORY, Changeset.ADD, None, -1),
                         next(changes))
        self.assertEqual(('tags', Node.DIRECTORY, Changeset.ADD, None, -1),
                         next(changes))
        self.assertEqual((u'tête', Node.DIRECTORY, Changeset.ADD, None, -1),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_file_edit(self):
        chgset = self.repos.get_changeset(3)
        self.assertEqual(3, chgset.rev)
        self.assertEqual('Fixed README.\n', chgset.message)
        self.assertEqual('kate', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 13, 24, 58, 234643, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual((u'tête/README.txt', Node.FILE, Changeset.EDIT,
                          u'tête/README.txt', 2), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_dir_moves(self):
        chgset = self.repos.get_changeset(5)
        self.assertEqual(5, chgset.rev)
        self.assertEqual('Moved directories.', chgset.message)
        self.assertEqual('kate', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 16, 25, 39, 658099, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual((u'tête/dir1/dir2', Node.DIRECTORY, Changeset.MOVE,
                          u'tête/dir2', 4), next(changes))
        self.assertEqual((u'tête/dir1/dir3', Node.DIRECTORY, Changeset.MOVE,
                          u'tête/dir3', 4), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_file_copy(self):
        chgset = self.repos.get_changeset(6)
        self.assertEqual(6, chgset.rev)
        self.assertEqual('More things to read', chgset.message)
        self.assertEqual('john', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 18, 56, 46, 985846, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual((u'tête/README2.txt', Node.FILE, Changeset.COPY,
                          u'tête/README.txt', 3), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_root_propset(self):
        chgset = self.repos.get_changeset(13)
        self.assertEqual(13, chgset.rev)
        self.assertEqual('Setting property on the repository_dir root',
                         chgset.message)
        changes = chgset.get_changes()
        self.assertEqual(('/', Node.DIRECTORY, Changeset.EDIT, '/', 12),
                         next(changes))
        self.assertEqual((u'tête', Node.DIRECTORY, Changeset.EDIT, u'tête', 6),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_base_path_rev(self):
        chgset = self.repos.get_changeset(9)
        self.assertEqual(9, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual(('branches/v1x/README.txt', Node.FILE,
                          Changeset.EDIT, u'tête/README.txt', 3),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_rename_and_edit(self):
        chgset = self.repos.get_changeset(14)
        self.assertEqual(14, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual((u'tête/README3.txt', Node.FILE,
                          Changeset.MOVE, u'tête/README2.txt', 13),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_edit_after_wc2wc_copy__original_deleted(self):
        chgset = self.repos.get_changeset(16)
        self.assertEqual(16, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual(('branches/v2', Node.DIRECTORY, Changeset.COPY,
                          'tags/v1.1', 14),
                         next(changes))
        self.assertEqual(('branches/v2/README2.txt', Node.FILE,
                          Changeset.EDIT, u'tête/README2.txt', 6),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_fancy_rename_double_delete(self):
        chgset = self.repos.get_changeset(19)
        self.assertEqual(19, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual((u'tête/mpp_proc', Node.DIRECTORY,
                          Changeset.MOVE, u'tête/Xprimary_proc', 18),
                         next(changes))
        self.assertEqual((u'tête/mpp_proc/Xprimary_pkg.vhd',
                          Node.FILE, Changeset.DELETE,
                          u'tête/Xprimary_proc/Xprimary_pkg.vhd', 18),
                         next(changes))
        self.assertEqual((u'tête/mpp_proc/Xprimary_proc', Node.DIRECTORY,
                          Changeset.COPY, u'tête/Xprimary_proc', 18),
                         next(changes))
        self.assertEqual((u'tête/mpp_proc/Xprimary_proc/Xprimary_pkg.vhd',
                          Node.FILE, Changeset.DELETE,
                          u'tête/Xprimary_proc/Xprimary_pkg.vhd', 18),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_copy_with_deletions_below_copy(self):
        """Regression test for #4900."""
        chgset = self.repos.get_changeset(22)
        self.assertEqual(22, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual((u'branches/v3', 'dir', 'copy',
                          u'tête', 21), next(changes))
        self.assertEqual((u'branches/v3/dir1', 'dir', 'delete',
                          u'tête/dir1', 21), next(changes))
        self.assertEqual((u'branches/v3/mpp_proc', 'dir', 'delete',
                          u'tête/mpp_proc', 21), next(changes))
        self.assertEqual((u'branches/v3/v2', 'dir', 'delete',
                          u'tête/v2', 21), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_utf_8(self):
        chgset = self.repos.get_changeset(20)
        self.assertEqual(20, chgset.rev)
        self.assertEqual(u'Chez moi ça marche\n', chgset.message)
        self.assertEqual(u'Jonas Borgström', chgset.author)

    def test_canonical_repos_path(self):
        # Assertion `svn_dirent_is_canonical` with leading double slashes
        # in repository path if os.name == 'posix' (#10390)
        DbRepositoryProvider(self.env).add_repository(
            'canonical-path', '//' + REPOS_PATH.lstrip('/'), 'direct-svnfs')
        repos = RepositoryManager(self.env).get_repository('canonical-path')
        self.assertEqual(REPOS_PATH, repos.path)

    if os.name != 'posix':
        del test_canonical_repos_path

    def test_merge_prop_renderer_without_deleted_branches(self):
        context = _create_context(self.env)
        context = context(self.repos.get_node('branches/v1x', HEAD).resource)
        renderer = svn_prop.SubversionMergePropertyRenderer(self.env)
        props = {'svn:mergeinfo': textwrap.dedent(u"""\
            /tête:1-20,23-26
            /branches/v3:22
            /branches/v2:16
            """)}
        result = renderer.render_property('svn:mergeinfo', 'browser',
                                          context, props)

        node = self._row_col(result, 1, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/branches/v2?rev=%d"'
                      % HEAD, node)
        self.assertIn('>/branches/v2</a>', node)
        node = self._row_col(result, 1, 2)
        self.assertIn(' title="16"', node)
        self.assertIn('>merged</a>', node)
        node = self._row_col(result, 1, 3)
        self.assertIn(' title="No revisions"', node)
        self.assertIn('>eligible</span>', node)

        node = self._row_col(result, 3, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/%s?rev=%d"'
                      % ('t%C3%AAte', HEAD), node)
        self.assertIn(u'>/tête</a>', node)
        node = self._row_col(result, 3, 2)
        self.assertIn(' title="1-20, 23-26"', node)
        self.assertIn(' href="/trac.cgi/log/repo/t%C3%AAte?revs=1-20%2C23-26"',
                      node)
        self.assertIn('>merged</a>', node)
        node = self._row_col(result, 3, 3)
        self.assertIn(' title="21"', node)
        self.assertIn(' href="/trac.cgi/changeset/21/repo/t%C3%AAte"', node)
        self.assertIn('>eligible</a>', node)

        self.assertNotIn('(toggle deleted branches)', unicode(result))
        self.assertNotIn('False', unicode(result))  # See #12125

    def test_merge_prop_renderer_with_deleted_branches(self):
        context = _create_context(self.env)
        context = context(self.repos.get_node('branches/v1x', HEAD).resource)
        renderer = svn_prop.SubversionMergePropertyRenderer(self.env)
        props = {'svn:mergeinfo': textwrap.dedent(u"""\
            /tête:19
            /branches/v3:22
            /branches/deleted:1,3-5,22
            """)}
        result = renderer.render_property('svn:mergeinfo', 'browser',
                                          context, props)

        node = self._row_col(result, 1, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/branches/v3?rev=%d"'
                      % HEAD, node)
        self.assertIn('>/branches/v3</a>', node)
        node = self._row_col(result, 1, 2)
        self.assertIn(' title="22"', node)
        self.assertIn('>merged</a>', node)
        node = self._row_col(result, 1, 3)
        self.assertIn(' title="No revisions"', node)
        self.assertIn('>eligible</span>', node)

        node = self._row_col(result, 2, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/%s?rev=%d"'
                      % ('t%C3%AAte', HEAD),
                      node)
        self.assertIn(u'>/tête</a>', node)
        node = self._row_col(result, 2, 2)
        self.assertIn(' title="19"', node)
        self.assertIn(' href="/trac.cgi/changeset/19/repo/t%C3%AAte"', node)
        self.assertIn('>merged</a>', node)
        node = self._row_col(result, 2, 3)
        self.assertIn(' title="13-14, 17-18, 20-21, 23-26"', node)
        self.assertIn(' href="/trac.cgi/log/repo/t%C3%AAte?revs='
                      '13-14%2C17-18%2C20-21%2C23-26"', node)
        self.assertIn('>eligible</a>', node)

        self.assertIn('(toggle deleted branches)', unicode(result))
        self.assertIn('<td>/branches/deleted</td>',
                      self._row_col(result, 3, 1))
        self.assertIn(u'<td colspan="2">1,\u200b3-5,\u200b22</td>',
                      self._row_col(result, 3, 2))

    def test_merge_prop_diff_renderer_added(self):
        context = _create_context(self.env)
        old_context = context(self.repos.get_node(u'tête', 20).resource)
        old_props = {'svn:mergeinfo': textwrap.dedent(u"""\
            /branches/v2:1,8-9,12-15
            /branches/v1x:12
            /branches/deleted:1,3-5,22
            """)}
        new_context = context(self.repos.get_node(u'tête', 21).resource)
        new_props = {'svn:mergeinfo': textwrap.dedent(u"""\
            /branches/v2:1,8-9,12-16
            /branches/v1x:12
            /branches/deleted:1,3-5,22
            """)}
        options = {}
        renderer = svn_prop.SubversionMergePropertyDiffRenderer(self.env)
        result = renderer.render_property_diff(
            'svn:mergeinfo', old_context, old_props, new_context,
            new_props, options)

        node = self._row_col(result, 1, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/branches/v2?rev=21"', node)
        self.assertIn('>/branches/v2</a>', node)
        node = self._row_col(result, 1, 2)
        self.assertIn(' title="16"', node)
        self.assertIn(' href="/trac.cgi/changeset/16/repo/branches/v2"', node)

    def test_merge_prop_diff_renderer_wrong_mergeinfo(self):
        rev = HEAD
        context = _create_context(self.env)
        old_context = context(self.repos.get_node(u'tête', rev - 1).resource)
        old_mergeinfo = '/missing:12\n'
        new_context = context(self.repos.get_node(u'tête', rev).resource)
        new_mergeinfo = '/missing:12-15\n'
        renderer = svn_prop.SubversionMergePropertyDiffRenderer(self.env)
        result = renderer.render_property_diff(
            'svn:mergeinfo', old_context, {'svn:mergeinfo': old_mergeinfo},
            new_context, {'svn:mergeinfo': new_mergeinfo}, {})

        node = self._row_col(result, 1, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/missing?rev=%d"'
                      % rev, node)
        self.assertIn('>/missing</a>', node)
        node = self._row_col(result, 1, 2)
        self.assertIn(' title="13-15"', node)
        self.assertIn(' href="/trac.cgi/log/repo/missing?revs=13-15"', node)

    def test_merge_prop_diff_renderer_added_svnmerge_integrated(self):
        """Property diff of svnmerge-integrated property (from
        svnmerge.py, used prior to svn 1.5) is rendered correctly.
        """
        context = _create_context(self.env)
        old_context = context(self.repos.get_node(u'tête', 20).resource)
        old_props = {'svnmerge-integrated': u"""\
        /branches/v2:1,8-9,12-15 /branches/v1x:12 /branches/deleted:1,3-5,22
        """}
        new_context = context(self.repos.get_node(u'tête', 21).resource)
        new_props = {'svnmerge-integrated': u"""\
        /branches/v2:1,8-9,12-16 /branches/v1x:12 /branches/deleted:1,3-5,22
        """}

        renderer = svn_prop.SubversionMergePropertyDiffRenderer(self.env)
        result = renderer.render_property_diff(
            'svnmerge-integrated', old_context, old_props, new_context,
            new_props, {})

        node = self._row_col(result, 1, 1)
        self.assertIn(' href="/trac.cgi/browser/repo/branches/v2?rev=21"',
                      node)
        self.assertIn('>/branches/v2</a>', node)
        node = self._row_col(result, 1, 2)
        self.assertIn(' title="16"', node)
        self.assertIn(' href="/trac.cgi/changeset/16/repo/branches/v2"',
                      node)

    def test_render_needslock(self):
        htdocs_location = 'http://assets.example.org/common'
        self.env.config.set('trac', 'htdocs_location', htdocs_location)
        context = _create_context(self.env)
        context.req.chrome['htdocs_location'] = htdocs_location
        context = context(self.repos.get_node(u'tête', HEAD).resource)
        renderer = svn_prop.SubversionPropertyRenderer(self.env)
        result = renderer.render_property('svn:needs-lock', None, context,
                                          {'svn:needs-lock': '*'})
        self.assertIn('src="http://assets.example.org/common/lock-locked.png"',
                      unicode(result))


class ScopedTests(object):

    def test_repos_normalize_path(self):
        self.assertEqual('/', self.repos.normalize_path('/'))
        self.assertEqual('/', self.repos.normalize_path(''))
        self.assertEqual('/', self.repos.normalize_path(None))
        self.assertEqual('dir1', self.repos.normalize_path('dir1'))
        self.assertEqual('dir1', self.repos.normalize_path('/dir1'))
        self.assertEqual('dir1', self.repos.normalize_path('dir1/'))
        self.assertEqual('dir1', self.repos.normalize_path('/dir1/'))

    def test_repos_normalize_rev(self):
        self.assertEqual(TETE, self.repos.normalize_rev('latest'))
        self.assertEqual(TETE, self.repos.normalize_rev('head'))
        self.assertEqual(TETE, self.repos.normalize_rev(''))
        self.assertEqual(TETE, self.repos.normalize_rev(None))
        self.assertEqual(5, self.repos.normalize_rev('5'))
        self.assertEqual(5, self.repos.normalize_rev(5))

    def test_rev_navigation(self):
        self.assertEqual(1, self.repos.oldest_rev)
        self.assertIsNone(self.repos.previous_rev(0))
        self.assertEqual(1, self.repos.previous_rev(2))
        self.assertEqual(TETE, self.repos.youngest_rev)
        self.assertEqual(2, self.repos.next_rev(1))
        self.assertEqual(3, self.repos.next_rev(2))
        # ...
        self.assertIsNone(self.repos.next_rev(TETE))

    def test_has_node(self):
        self.assertFalse(self.repos.has_node('/dir1', 3))
        self.assertTrue(self.repos.has_node('/dir1', 4))

    def test_get_node(self):
        node = self.repos.get_node('/dir1')
        self.assertEqual('dir1', node.name)
        self.assertEqual('/dir1', node.path)
        self.assertEqual(Node.DIRECTORY, node.kind)
        self.assertEqual(TETE, node.rev)
        self.assertEqual(5, node.created_rev)
        self.assertEqual(datetime(2005, 4, 1, 16, 25, 39, 658099, utc),
                         node.last_modified)
        node = self.repos.get_node('/README.txt')
        self.assertEqual('README.txt', node.name)
        self.assertEqual('/README.txt', node.path)
        self.assertEqual(Node.FILE, node.kind)
        self.assertEqual(TETE, node.rev)
        self.assertEqual(3, node.created_rev)
        self.assertEqual(datetime(2005, 4, 1, 13, 24, 58, 234643, utc),
                         node.last_modified)

    def test_get_node_specific_rev(self):
        node = self.repos.get_node('/dir1', 4)
        self.assertEqual('dir1', node.name)
        self.assertEqual('/dir1', node.path)
        self.assertEqual(Node.DIRECTORY, node.kind)
        self.assertEqual(4, node.rev)
        self.assertEqual(datetime(2005, 4, 1, 15, 42, 35, 450595, utc),
                         node.last_modified)
        node = self.repos.get_node('/README.txt', 2)
        self.assertEqual('README.txt', node.name)
        self.assertEqual('/README.txt', node.path)
        self.assertEqual(Node.FILE, node.kind)
        self.assertEqual(2, node.rev)
        self.assertEqual(datetime(2005, 4, 1, 13, 12, 18, 216267, utc),
                         node.last_modified)

    def test_get_dir_entries(self):
        node = self.repos.get_node('/')
        entries = node.get_entries()
        self.assertEqual('dir1', next(entries).name)
        self.assertEqual('mpp_proc', next(entries).name)
        self.assertEqual('v2', next(entries).name)
        self.assertEqual('README3.txt', next(entries).name)
        self.assertEqual(u'R\xe9sum\xe9.txt', next(entries).name)
        self.assertEqual('README.txt', next(entries).name)
        self.assertRaises(StopIteration, next, entries)

    def test_get_file_entries(self):
        node = self.repos.get_node('/README.txt')
        entries = node.get_entries()
        self.assertRaises(StopIteration, next, entries)

    def test_get_dir_content(self):
        node = self.repos.get_node('/dir1')
        self.assertIsNone(node.content_length)
        self.assertIsNone(node.content_type)
        self.assertIsNone(node.get_content())

    def test_get_file_content(self):
        node = self.repos.get_node('/README.txt')
        self.assertEqual(8, node.content_length)
        self.assertEqual('text/plain', node.content_type)
        self.assertEqual('A test.\n', node.get_content().read())

    def test_get_dir_properties(self):
        f = self.repos.get_node('/dir1')
        props = f.get_properties()
        self.assertEqual(0, len(props))

    def test_get_file_properties(self):
        f = self.repos.get_node('/README.txt')
        props = f.get_properties()
        self.assertEqual('native', props['svn:eol-style'])
        self.assertEqual('text/plain', props['svn:mime-type'])

    # Revision Log / node history

    def test_get_history_scope(self):
        """Regression test for #9504"""
        node = self.repos.get_node('/')
        history = list(node.get_history())
        self.assertEqual(('/', 1, 'add'), history[-1])
        initial_cset = self.repos.get_changeset(history[-1][1])
        self.assertEqual(1, initial_cset.rev)

    def test_get_node_history(self):
        node = self.repos.get_node('/README3.txt')
        history = node.get_history()
        self.assertEqual(('README3.txt', 14, 'copy'), next(history))
        self.assertEqual(('README2.txt', 6, 'copy'), next(history))
        self.assertEqual(('README.txt', 3, 'edit'), next(history))
        self.assertEqual(('README.txt', 2, 'add'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_node_history_follow_copy(self):
        node = self.repos.get_node('dir1/dir3', )
        history = node.get_history()
        self.assertEqual(('dir1/dir3', 5, 'copy'), next(history))
        self.assertEqual(('dir3', 4, 'add'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_copy_ancestry(self):
        node = self.repos.get_node(u'/README3.txt')
        ancestry = node.get_copy_ancestry()
        self.assertEqual([(u'README2.txt', 13),
                          (u'README.txt', 3)], ancestry)
        for path, rev in ancestry:
            self.repos.get_node(path, rev)  # shouldn't raise NoSuchNode

    def test_get_copy_ancestry_for_move(self):
        node = self.repos.get_node(u'/dir1/dir2', 5)
        ancestry = node.get_copy_ancestry()
        self.assertEqual([(u'dir2', 4)], ancestry)
        for path, rev in ancestry:
            self.repos.get_node(path, rev)  # shouldn't raise NoSuchNode

    def test_get_branch_origin(self):
        node = self.repos.get_node(u'/README3.txt')
        self.assertEqual(14, node.get_branch_origin())
        node = self.repos.get_node(u'/dir1/dir2', 5)
        self.assertEqual(5, node.get_branch_origin())

    # Revision Log / path history

    def test_get_path_history(self):
        history = self.repos.get_path_history('dir3', None)
        self.assertEqual(('dir3', 5, 'delete'), next(history))
        self.assertEqual(('dir3', 4, 'add'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_path_history_copied_file(self):
        history = self.repos.get_path_history('README3.txt', None)
        self.assertEqual(('README3.txt', 14, 'copy'), next(history))
        self.assertEqual(('README2.txt', 6, 'unknown'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_get_path_history_copied_dir(self):
        history = self.repos.get_path_history('dir1/dir3', None)
        self.assertEqual(('dir1/dir3', 5, 'copy'), next(history))
        self.assertEqual(('dir3', 4, 'unknown'), next(history))
        self.assertRaises(StopIteration, next, history)

    def test_changeset_repos_creation(self):
        chgset = self.repos.get_changeset(0)
        self.assertEqual(0, chgset.rev)
        self.assertEqual('', chgset.message)
        self.assertEqual('', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 9, 57, 41, 312767, utc),
                         chgset.date)
        self.assertRaises(StopIteration, next, chgset.get_changes())

    def test_changeset_added_dirs(self):
        chgset = self.repos.get_changeset(4)
        self.assertEqual(4, chgset.rev)
        self.assertEqual('More directories.', chgset.message)
        self.assertEqual('john', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 15, 42, 35, 450595, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual(('dir1', Node.DIRECTORY, 'add', None, -1),
                         next(changes))
        self.assertEqual(('dir2', Node.DIRECTORY, 'add', None, -1),
                         next(changes))
        self.assertEqual(('dir3', Node.DIRECTORY, 'add', None, -1),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_file_edit(self):
        chgset = self.repos.get_changeset(3)
        self.assertEqual(3, chgset.rev)
        self.assertEqual('Fixed README.\n', chgset.message)
        self.assertEqual('kate', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 13, 24, 58, 234643, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual(('README.txt', Node.FILE, Changeset.EDIT,
                          'README.txt', 2), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_dir_moves(self):
        chgset = self.repos.get_changeset(5)
        self.assertEqual(5, chgset.rev)
        self.assertEqual('Moved directories.', chgset.message)
        self.assertEqual('kate', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 16, 25, 39, 658099, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual(('dir1/dir2', Node.DIRECTORY, Changeset.MOVE,
                          'dir2', 4), next(changes))
        self.assertEqual(('dir1/dir3', Node.DIRECTORY, Changeset.MOVE,
                          'dir3', 4), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_file_copy(self):
        chgset = self.repos.get_changeset(6)
        self.assertEqual(6, chgset.rev)
        self.assertEqual('More things to read', chgset.message)
        self.assertEqual('john', chgset.author)
        self.assertEqual(datetime(2005, 4, 1, 18, 56, 46, 985846, utc),
                         chgset.date)

        changes = chgset.get_changes()
        self.assertEqual(('README2.txt', Node.FILE, Changeset.COPY,
                          'README.txt', 3), next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_root_propset(self):
        chgset = self.repos.get_changeset(13)
        self.assertEqual(13, chgset.rev)
        self.assertEqual('Setting property on the repository_dir root',
                         chgset.message)
        changes = chgset.get_changes()
        self.assertEqual(('/', Node.DIRECTORY, Changeset.EDIT, '/', 6),
                         next(changes))
        self.assertRaises(StopIteration, next, changes)

    def test_changeset_copy_from_outside_and_delete(self):
        chgset = self.repos.get_changeset(21)
        self.assertEqual(21, chgset.rev)
        self.assertEqual('copy from outside of the scope + delete',
                         chgset.message)
        changes = chgset.get_changes()
        self.assertEqual(('v2', Node.DIRECTORY, Changeset.ADD, None, -1),
                         next(changes))
        self.assertEqual(('v2/README2.txt', Node.FILE, Changeset.DELETE,
                          None, -1), next(changes))
        self.assertEqual(('v2/dir1', Node.DIRECTORY, Changeset.DELETE,
                          None, -1), next(changes))
        self.assertRaises(StopIteration, next, changes)


class RecentPathScopedTests(object):

    def test_rev_navigation(self):
        self.assertFalse(self.repos.has_node('/', 1))
        self.assertFalse(self.repos.has_node('/', 2))
        self.assertFalse(self.repos.has_node('/', 3))
        self.assertTrue(self.repos.has_node('/', 4))
        # We can't make this work anymore because of #5213.
        # self.assertEqual(4, self.repos.oldest_rev)
        self.assertEqual(1, self.repos.oldest_rev)  # should really be 4...
        self.assertIsNone(self.repos.previous_rev(4))


class NonSelfContainedScopedTests(object):

    def test_mixed_changeset(self):
        chgset = self.repos.get_changeset(7)
        self.assertEqual(7, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual(('/', Node.DIRECTORY, Changeset.ADD, None, -1),
                         next(changes))
        self.assertRaises(TracError, lambda: self.repos.get_node(None, 6))


class AnotherNonSelfContainedScopedTests(object):

    def test_mixed_changeset_with_edit(self):
        chgset = self.repos.get_changeset(9)
        self.assertEqual(9, chgset.rev)
        changes = chgset.get_changes()
        self.assertEqual(('v1x/README.txt', Node.FILE, Changeset.EDIT,
                          'v1x/README.txt', 8),
                         next(changes))


class ExternalsPropertyTests(object):

    def _modify_repository(self, reponame, changes):
        DbRepositoryProvider(self.env).modify_repository(reponame, changes)

    def _set_tracini_externals(self, *values):
        config = self.env.config
        for idx, value in enumerate(values):
            config.set('svn:externals', str(idx), value)

    def _render(self, *values):
        props = {'svn:externals': '\n'.join(values)}
        renderer = svn_prop.SubversionPropertyRenderer(self.env)
        context = RenderingContext(Resource('repository', REPOS_NAME)
                                   .child('source', 'build/posix', 42))
        return renderer.render_property('svn:externals', None, context, props)

    def _parse_result(self, ul):
        items = []
        for li in ul.children:
            if li.tag != 'li':
                continue
            for a in li.children:
                if a.tag != 'a':
                    continue
                href = a.attrib.get('href')
                title = a.attrib.get('title')
                text = Fragment.__unicode__(a)
                items.append({'text()': text, '@href': href, '@title': title})
        return items

    def test_match_property(self):
        renderer = svn_prop.SubversionPropertyRenderer(self.env)
        rv = renderer.match_property('svn:externals', None)
        self.assertTrue(1 <= rv < 10000)

    def test_render_property_without_tracini(self):
        result = self._parse_result(self._render(
            'blah svn://server/repos1',
            'vendor http://example.org/svn/eng-soft'))
        self.assertEqual(
            {'text()': u'blah', '@href': None,
             '@title': Markup('No svn:externals configured in trac.ini')},
                         result[0])
        self.assertEqual(
            {'text()': u'vendor',
             '@href': Markup('http://example.org/svn/eng-soft'),
             # XXX           v should be "//"
             '@title': Markup('http:/example.org/svn/eng-soft')},
            result[1])
        self.assertEqual(2, len(result))

    def test_render_property_non_absolute_url(self):
        externals = [u'blah1 ../src',
                     u'blah2 ^/src',
                     u'blah3 /svn/trunk',
                     u'blah4 //localhost/svn']
        result = self._parse_result(self._render(*externals))
        self.assertEqual(
            [{'text()': externals[0], '@href': None, '@title': Markup()},
             {'text()': externals[1], '@href': None, '@title': Markup()},
             {'text()': externals[2], '@href': None, '@title': Markup()},
             {'text()': externals[3], '@href': None, '@title': Markup()}],
            result)

    def test_render_property_comment(self):
        result = self._parse_result(self._render(
            '   # For blah',
            'blah svn://server/repos1',
            '',
            '   # path rev url .....',
            'vendor http://example.org/svn/eng-soft'))
        self.assertEqual(
            {'text()': u'   # For blah', '@href': None, '@title':  Markup()},
            result[0])
        self.assertEqual(u'blah', result[1]['text()'])
        self.assertEqual(
            {'text()': u'   # path rev url .....', '@href': None,
             '@title': Markup()},
            result[2])
        self.assertEqual(u'vendor', result[3]['text()'])
        self.assertEqual(4, len(result))

    def test_render_property_with_tracini(self):
        self._set_tracini_externals(
            'svn://server/repos1 http://trac/proj1/browser/$path?rev=$rev',
            'http://example.org/svn/eng-soft '
                'http://example.org/trac/eng-soft/browser/$path?rev=$rev')
        result = self._parse_result(self._render(
            'blah          svn://server/repos1',
            'blah-doc      svn://server/repos1/doc',
            'blah-r42 -r42 svn://server/repos1/branches/1.0-stable',
            'vendor        http://example.org/svn/eng-soft'))
        self.assertEqual({'text()': 'blah in svn://server/repos1',
                          '@href': 'http://trac/proj1/browser/?rev=',
                          '@title': ' in svn://server/repos1'}, result[0])
        self.assertEqual({'text()': 'blah-doc in svn://server/repos1',
                          '@href': 'http://trac/proj1/browser/doc?rev=',
                          '@title': 'doc in svn://server/repos1'}, result[1])
        self.assertEqual({'text()': 'blah-r42 at revision 42 in '
                                    'svn://server/repos1',
                          '@href': 'http://trac/proj1/browser/branches/'
                                   '1.0-stable?rev=42',
                          '@title': 'branches/1.0-stable at revision 42 in '
                                    'svn://server/repos1'}, result[2])
        self.assertEqual({
            'text()': 'vendor in http://example.org/svn/eng-soft',
            '@href': 'http://example.org/trac/eng-soft/browser/?rev=',
            '@title': ' in http://example.org/svn/eng-soft'}, result[3])
        self.assertEqual(4, len(result))


class SubversionRepositoryTestCase(unittest.TestCase):
    """Test cases for SubversionRepository."""

    def setUp(self):
        self.env = EnvironmentStub()
        dbprovider = DbRepositoryProvider(self.env)
        dbprovider.add_repository(REPOS_NAME, self.path, 'direct-svnfs')
        dbprovider.modify_repository(REPOS_NAME, {'url': URL})
        self.repos = RepositoryManager(self.env).get_repository(REPOS_NAME)

    def tearDown(self):
        self.repos.close()
        self.repos = None
        # clear cached repositories to avoid TypeError on termination (#11505)
        RepositoryManager(self.env).reload_repositories()
        self.env.reset_db()
        # needed to avoid issue with 'WindowsError: The process cannot access
        # the file ... being used by another process: ...\rep-cache.db'
        self.env.shutdown(get_thread_id())


class SvnCachedRepositoryTestCase(unittest.TestCase):
    """Test cases for SvnCachedRepository."""

    def setUp(self):
        self.env = EnvironmentStub()
        dbprovider = DbRepositoryProvider(self.env)
        dbprovider.add_repository(REPOS_NAME, self.path, 'svn')
        dbprovider.modify_repository(REPOS_NAME, {'url': URL})
        self.repos = RepositoryManager(self.env).get_repository(REPOS_NAME)
        self.repos.sync()

    def tearDown(self):
        self.env.reset_db()
        self.repos.close()
        self.repos = None
        # clear cached repositories to avoid TypeError on termination (#11505)
        RepositoryManager(self.env).reload_repositories()


class SubversionConnectorTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub()
        self.repos_path = mkdtemp()
        self.dbprovider = DbRepositoryProvider(self.env)
        pool = core.svn_pool_create(None)
        repos.svn_repos_create(self.repos_path, '', '', None, None, pool)
        self.dbprovider.add_repository(REPOS_NAME, self.repos_path, 'svn')

    def tearDown(self):
        self.env.reset_db()
        # clear cached repositories to avoid TypeError on termination (#11505)
        RepositoryManager(self.env).reload_repositories()
        repos.svn_repos_delete(self.repos_path)

    def _svn_version_from_system_info(self):
        svn_version = None
        for name, version in self.env.system_info:
            if name == 'Subversion':
                svn_version = version
        return svn_version

    def test_get_system_info(self):
        self.assertIsNotNone(self._svn_version_from_system_info())


def test_suite():
    global REPOS_PATH
    suite = unittest.TestSuite()
    if has_svn:
        REPOS_PATH = mkdtemp()
        os.rmdir(REPOS_PATH)
        tests = [(NormalTests, ''),
                 (ScopedTests, u'/tête'),
                 (RecentPathScopedTests, u'/tête/dir1'),
                 (NonSelfContainedScopedTests, '/tags/v1'),
                 (AnotherNonSelfContainedScopedTests, '/branches'),
                 (ExternalsPropertyTests, ''),
                 ]
        skipped = {
            'SvnCachedRepositoryNormalTests': [
                'test_changeset_repos_creation',
                ],
            'SvnCachedRepositoryScopedTests': [
                'test_changeset_repos_creation',
                'test_rev_navigation',
                ],
            }
        for test, scope in tests:
            tc = type('SubversionRepository' + test.__name__,
                      (SubversionRepositoryTestCase, test),
                      {'path': REPOS_PATH + scope})
            suite.addTest(unittest.makeSuite(
                tc, suiteClass=SubversionRepositoryTestSetup))
            tc = type('SvnCachedRepository' + test.__name__,
                      (SvnCachedRepositoryTestCase, test),
                      {'path': REPOS_PATH + scope})
            for skip in skipped.get(tc.__name__, []):
                setattr(tc, skip, lambda self: None)  # no skip, so we cheat...
            suite.addTest(unittest.makeSuite(
                tc, suiteClass=SubversionRepositoryTestSetup))
        suite.addTest(unittest.makeSuite(SubversionConnectorTestCase))
    else:
        print("SKIP: tracopt/versioncontrol/svn/tests/svn_fs.py (no svn "
              "bindings)")
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import unittest

from trac.test import EnvironmentStub, MockRequest
from trac.mimeview.rst import ReStructuredTextRenderer, has_docutils
from trac.web.chrome import web_context


class ReStructuredTextRendererTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=[ReStructuredTextRenderer])
        self.renderer = ReStructuredTextRenderer(self.env)
        self.req = MockRequest(self.env)
        self.context = web_context(self.req)

    def tearDown(self):
        self.env.reset_db()

    def _render(self, text):
        return self.renderer.render(self.context, 'rst', text)

    _rst_text = """
{{{#!rst
`javascript link`_.
`call link`_.
`relative link`_.
`normal link`_.

.. _javascript link: javascript:alert(42)
.. _call link: call:01234567890
.. _relative link: /path/to/index.html
.. _normal link: https://example.org/
}}}
"""

    def test_safe_schemes(self):
        result = self._render(self._rst_text)
        self.assertNotIn(' href="javascript:alert(42)"', result)
        self.assertNotIn(' href="call:01234567890"', result)
        self.assertIn(' href="/path/to/index.html"', result)
        self.assertIn(' href="https://example.org/"', result)

    def test_call_in_safe_schemes(self):
        self.env.config.set('wiki', 'safe_schemes', 'call')
        result = self._render(self._rst_text)
        self.assertNotIn(' href="javascript:alert(42)"', result)
        self.assertIn(' href="call:01234567890"', result)
        self.assertIn(' href="/path/to/index.html"', result)
        self.assertNotIn(' href="https://example.org/"', result)

    def test_render_unsafe_content(self):
        self.env.config.set('wiki', 'render_unsafe_content', 'enabled')
        self.env.config.set('wiki', 'safe_schemes', '')
        result = self._render(self._rst_text)
        self.assertIn(' href="javascript:alert(42)"', result)
        self.assertIn(' href="call:01234567890"', result)
        self.assertIn(' href="/path/to/index.html"', result)
        self.assertIn(' href="https://example.org/"', result)

    def test_cross_origin_images(self):
        def test (directive):
            result = self._render("""
.. %(directive)s:: https://example.org/foo.png
   :alt:
.. %(directive)s:: //example.net/foo.png
   :alt:
.. %(directive)s:: /path/to/foo.png
   :alt:
.. %(directive)s:: ./foo.png
   :alt:
.. %(directive)s:: foo.png
   :alt:
.. %(directive)s:: data:image/png,foo
   :alt:
""" % {'directive': directive})
            self.assertIn('<img crossorigin="anonymous" alt="" '
                          'src="https://example.org/foo.png" />', result)
            self.assertIn('<img crossorigin="anonymous" alt="" '
                          'src="//example.net/foo.png" />', result)
            self.assertIn('<img alt="" src="/path/to/foo.png" />', result)
            self.assertIn('<img alt="" src="./foo.png" />', result)
            self.assertIn('<img alt="" src="foo.png" />', result)
            self.assertIn('<img alt="" src="data:image/png,foo" />', result)

        test('image')
        test('figure')


def test_suite():
    suite = unittest.TestSuite()
    if has_docutils:
        suite.addTest(unittest.makeSuite(ReStructuredTextRendererTestCase))
    else:
        print('SKIP: mimeview/tests/rst (no docutils installed)')
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 Edgewall Software
# Copyright (C) 2012 Franz Mayer <franz.mayer@gefasoft.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/.

from trac.core import Component, implements
from trac.resource import ResourceNotFound
from trac.ticket.api import ITicketActionController
from trac.ticket.default_workflow import ConfigurableTicketWorkflow
from trac.ticket.model import Milestone
from trac.util import to_list
from trac.util.translation import _
from trac.web.chrome import add_warning

revision = "$Rev$"
url = "$URL$"


class MilestoneOperation(Component):
    """Sets milestone for specified resolutions.

    Example:
    {{{#!ini
    [ticket-workflow]
    resolve.operations = set_resolution,set_milestone
    resolve.milestone = invalid,wontfix,duplicate,worksforme -> rejected
    }}}

    When setting resolution to `duplicate` the milestone will
    automatically change to `rejected`. If user changes milestone
    manually when resolving the ticket, this workflow operation has
    ''no effect''.

    Don't forget to add `MilestoneOperation` to the `workflow` option
    in the `[ticket]` section of TracIni. When added to the default
    value of `workflow`, the line will look like this:
    {{{#!ini
    [ticket]
    workflow = ConfigurableTicketWorkflow,MilestoneOperation
    }}}
    """

    implements(ITicketActionController)

    def get_ticket_actions(self, req, ticket):
        controller = ConfigurableTicketWorkflow(self.env)
        return controller.get_actions_by_operation_for_req(req, ticket,
                                                           'set_milestone')

    def get_all_status(self):
        return []

    def render_ticket_action_control(self, req, ticket, action):
        actions = ConfigurableTicketWorkflow(self.env).actions
        label = actions[action]['label']
        hint = None
        old_milestone = ticket._old.get('milestone')
        if old_milestone is None:
            resolutions, milestone = \
                self._get_resolutions_and_milestone(action)
            if resolutions:
                try:
                    Milestone(self.env, milestone)
                except ResourceNotFound:
                    pass
                else:
                    res_hint = ', '.join("'%s'" % r for r in resolutions)
                    hint = _("For resolution %(resolutions)s the milestone "
                             "will be set to '%(milestone)s'.",
                             resolutions=res_hint, milestone=milestone)
        return label, None, hint

    def get_ticket_changes(self, req, ticket, action):
        old_milestone = ticket._old.get('milestone')
        if old_milestone is None:
            new_milestone = self._get_resolutions_and_milestone(action)[1]
            try:
                Milestone(self.env, new_milestone)
            except ResourceNotFound:
                add_warning(req, _("Milestone %(name)s does not exist.",
                                   name=new_milestone))
            else:
                self.log.info("Changed milestone from %s to %s",
                              old_milestone, new_milestone)
                return {'milestone': new_milestone}
        return {}

    def apply_action_side_effects(self, req, ticket, action):
        pass

    def _get_resolutions_and_milestone(self, action):
        transitions = self.config.get('ticket-workflow', action + '.milestone')
        milestone = None
        resolutions = []
        try:
            transition = to_list(transitions, sep='->')
        except ValueError:
            pass
        else:
            if len(transition) == 2:
                resolutions = to_list(transition[0])
                milestone = transition[1]
        return resolutions, milestone



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import unittest
from datetime import timedelta

from trac.test import locale_en
from trac.ticket.query import QueryModule
from trac.ticket.report import ReportModule
from trac.ticket.roadmap import RoadmapModule
from trac.ticket.model import Milestone
from trac.ticket.test import insert_ticket
from trac.util.datefmt import (datetime_now, format_datetime, pretty_timedelta,
                               utc)
from trac.wiki.tests import formatter


TICKET_TEST_CASES = u"""
============================== ticket: link resolver
bug:1
issue:1
ticket:1
ticket:12
ticket:abc
------------------------------
<p>
<a class="new ticket" href="/ticket/1" title="#1: This is the summary (new)">bug:1</a>
<a class="new ticket" href="/ticket/1" title="#1: This is the summary (new)">issue:1</a>
<a class="new ticket" href="/ticket/1" title="#1: This is the summary (new)">ticket:1</a>
<a class="missing ticket">ticket:12</a>
<a class="missing ticket">ticket:abc</a>
</p>
------------------------------
============================== ticket: link resolver + arguments
bug:1#comment:3
issue:1#comment:3
ticket:1#comment:3
ticket:1?format=csv
------------------------------
<p>
<a class="new ticket" href="/ticket/1#comment:3" title="#1: This is the summary (new)">bug:1#comment:3</a>
<a class="new ticket" href="/ticket/1#comment:3" title="#1: This is the summary (new)">issue:1#comment:3</a>
<a class="new ticket" href="/ticket/1#comment:3" title="#1: This is the summary (new)">ticket:1#comment:3</a>
<a class="new ticket" href="/ticket/1?format=csv" title="#1: This is the summary (new)">ticket:1?format=csv</a>
</p>
------------------------------
============================== ticket: link resolver with ranges
bug:12-14,33
issue:12-14,33
ticket:12-14,33
ticket:12,33?order=created
------------------------------
<p>
<a href="/query?id=12-14%2C33" title="Tickets 12-14, 33">bug:12-14,33</a>
<a href="/query?id=12-14%2C33" title="Tickets 12-14, 33">issue:12-14,33</a>
<a href="/query?id=12-14%2C33" title="Tickets 12-14, 33">ticket:12-14,33</a>
<a href="/query?id=12%2C33&amp;order=created" title="Tickets 12, 33">ticket:12,33?order=created</a>
</p>
------------------------------
============================== ticket link shorthand form
#1, #2
#12, #abc
------------------------------
<p>
<a class="new ticket" href="/ticket/1" title="#1: This is the summary (new)">#1</a>, <a class="missing ticket">#2</a>
<a class="missing ticket">#12</a>, #abc
</p>
------------------------------
============================== ticket link shorthand form with ranges
#1-5,42
#1,3,5,7
------------------------------
<p>
<a href="/query?id=1-5%2C42" title="Tickets 1-5, 42">#1-5,42</a>
<a href="/query?id=1%2C3%2C5%2C7" title="Tickets 1, 3, 5, 7">#1,3,5,7</a>
</p>
------------------------------
============================== ticket link shorthand form with long ranges (#10111 regression)
#1-123456789012345678901234
------------------------------
<p>
<a href="/query?id=1-123456789012345678901234" title="Tickets 1-123456789012345678901234">#1-123456789012345678901234</a>
</p>
------------------------------
============================== escaping the above
!#1
------------------------------
<p>
#1
</p>
------------------------------
#1
============================== InterTrac for tickets
trac:ticket:2041
[trac:ticket:2041 Trac #2041]
#T2041
#trac2041
------------------------------
<p>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/ticket%3A2041" title="ticket:2041 in The Trac Project"><span class="icon"></span>trac:ticket:2041</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/ticket%3A2041" title="ticket:2041 in The Trac Project"><span class="icon"></span>Trac #2041</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/ticket%3A2041" title="ticket:2041 in The Trac Project"><span class="icon"></span>#T2041</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/ticket%3A2041" title="ticket:2041 in The Trac Project"><span class="icon"></span>#trac2041</a>
</p>
------------------------------
============================== Ticket InterTrac shorthands
T:#2041
trac:#2041
------------------------------
<p>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/%232041" title="#2041 in The Trac Project"><span class="icon"></span>T:#2041</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/%232041" title="#2041 in The Trac Project"><span class="icon"></span>trac:#2041</a>
</p>
------------------------------
============================== ticket syntax with unicode digits
#⁴²
#1-⁵,42
#1,³,5,7
#T²⁰⁴¹
#trac²⁰⁴¹
------------------------------
<p>
#⁴²
<a class="new ticket" href="/ticket/1" title="#1: This is the summary (new)">#1</a>-⁵,42
<a class="new ticket" href="/ticket/1" title="#1: This is the summary (new)">#1</a>,³,5,7
#T²⁰⁴¹
#trac²⁰⁴¹
</p>
------------------------------
""" # "


def ticket_setup(tc):
    config = tc.env.config
    config.set('ticket-custom', 'custom1', 'text')
    config.save()
    insert_ticket(tc.env, reporter='santa', summary='This is the summary',
                  status='new')

def ticket_teardown(tc):
    config = tc.env.config
    config.remove('ticket-custom', 'custom1')
    config.save()
    tc.env.reset_db()



REPORT_TEST_CASES = u"""
============================== report link shorthand form
{1}, {2}
{12}, {abc}
------------------------------
<p>
<a class="report" href="/report/1">{1}</a>, <a class="report" href="/report/2">{2}</a>
<a class="missing report" title="report does not exist">{12}</a>, {abc}
</p>
------------------------------
============================== escaping the above
!{1}
------------------------------
<p>
{1}
</p>
------------------------------
{1}
============================== ticket shorthands, not numerical HTML entities
&#1; &#23;
------------------------------
<p>
&amp;#1; &amp;#23;
</p>
------------------------------
&amp;#1; &amp;#23;
============================== report link with non-digits
report:blah
------------------------------
<p>
<a class="missing report" title="report does not exist">report:blah</a>
</p>
------------------------------
<a class="missing report" title="report does not exist">report:blah</a>
============================== InterTrac for reports
trac:report:1
[trac:report:1 Trac r1]
{T1}
{trac1}
{trac 1}
------------------------------
<p>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/report%3A1" title="report:1 in The Trac Project"><span class="icon"></span>trac:report:1</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/report%3A1" title="report:1 in The Trac Project"><span class="icon"></span>Trac r1</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/report%3A1" title="report:1 in The Trac Project"><span class="icon"></span>{T1}</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/report%3A1" title="report:1 in The Trac Project"><span class="icon"></span>{trac1}</a>
<a class="ext-link" href="https://trac.edgewall.org/intertrac/report%3A1" title="report:1 in The Trac Project"><span class="icon"></span>{trac 1}</a>
</p>
------------------------------
============================== report syntax with unicode digits
{⁴²} !{⁴²}
{T⁴²}
{trac⁴²}
------------------------------
<p>
{⁴²} !{⁴²}
{T⁴²}
{trac⁴²}
</p>
------------------------------
""" # '

def report_setup(tc):
    def create_report(tc, id):
        tc.env.db_transaction("""
            INSERT INTO report (id,title,query,description)
            VALUES (%s,%s,'SELECT 1','')""", (id, 'Report %s' % id))
    create_report(tc, 1)
    create_report(tc, 2)


dt_past = datetime_now(utc) - timedelta(days=1)
dt_future = datetime_now(utc) + timedelta(days=1)


MILESTONE_TEST_CASES = u"""
============================== milestone: link resolver
milestone:foo
[milestone:boo Milestone Boo]
[milestone:roo Milestone Roo]
[milestone:woo Milestone Woo]
[milestone:zoo Milestone Zoo]
------------------------------
<p>
<a class="missing milestone" href="/milestone/foo" rel="nofollow">milestone:foo</a>
<a class="milestone" href="/milestone/boo" title="No date set">Milestone Boo</a>
<a class="closed milestone" href="/milestone/roo" title="Completed %(dt_past)s ago (%(datestr_past)s)">Milestone Roo</a>
<a class="milestone" href="/milestone/woo" title="Due in %(dt_future)s (%(datestr_future)s)">Milestone Woo</a>
<a class="milestone" href="/milestone/zoo" title="%(dt_past)s late (%(datestr_past)s)">Milestone Zoo</a>
</p>
------------------------------
============================== milestone: link resolver + arguments
milestone:?action=new
[milestone:boo#KnownIssues Known Issues for 1.0]
------------------------------
<p>
<a class="missing milestone" href="/milestone/?action=new" rel="nofollow">milestone:?action=new</a>
<a class="milestone" href="/milestone/boo#KnownIssues" title="No date set">Known Issues for 1.0</a>
</p>
------------------------------
""" % {'dt_past': pretty_timedelta(dt_past),
       'dt_future': pretty_timedelta(dt_future),
       'datestr_past': format_datetime(dt_past, locale=locale_en, tzinfo=utc),
       'datestr_future': format_datetime(dt_future, locale=locale_en,
                                         tzinfo=utc)} #"

def milestone_setup(tc):
    boo = Milestone(tc.env)
    boo.name = 'boo'
    boo.completed = boo.due = None
    boo.insert()
    roo = Milestone(tc.env)
    roo.name = 'roo'
    roo.completed = dt_past
    roo.due = None
    roo.insert()
    woo = Milestone(tc.env)
    woo.name = 'woo'
    woo.completed = None
    woo.due = dt_future
    woo.insert()
    zoo = Milestone(tc.env)
    zoo.name = 'zoo'
    zoo.completed = None
    zoo.due = dt_past
    zoo.insert()

def milestone_teardown(tc):
    tc.env.reset_db()



QUERY_TEST_CASES = u"""
============================== query: link resolver
query:?order=priority

query:?order=priority&owner=me

query:?type=résumé

query:status=new|reopened

query:reporter!=

query:reporter=joe|jack&owner=me

query:group=owner

query:verbose=1

query:summary=résumé
------------------------------
<p>
<a class="query" href="/query?order=priority">query:?order=priority</a>
</p>
<p>
<a class="query" href="/query?order=priority&amp;owner=me">query:?order=priority&amp;owner=me</a>
</p>
<p>
<a class="query" href="/query?type=r%C3%A9sum%C3%A9">query:?type=résumé</a>
</p>
<p>
<a class="query" href="/query?status=new&amp;status=reopened&amp;order=priority">query:status=new|reopened</a>
</p>
<p>
<a class="query" href="/query?reporter=!&amp;order=priority">query:reporter!=</a>
</p>
<p>
<a class="query" href="/query?owner=me&amp;reporter=joe&amp;reporter=jack&amp;order=priority">query:reporter=joe|jack&amp;owner=me</a>
</p>
<p>
<a class="query" href="/query?group=owner&amp;order=priority">query:group=owner</a>
</p>
<p>
<a class="query" href="/query?order=priority&amp;row=description">query:verbose=1</a>
</p>
<p>
<a class="query" href="/query?summary=r%C3%A9sum%C3%A9&amp;order=priority">query:summary=résumé</a>
</p>
------------------------------
============================== TicketQuery macro: no results, list form
Reopened tickets: [[TicketQuery(status=reopened)]]
------------------------------
<p>
Reopened tickets: <span class="query_no_results">No results</span>
</p>
------------------------------
============================== TicketQuery macro: no results, count 0 (raw)
Reopened tickets: [[TicketQuery(status=reopened, format=rawcount)]]
------------------------------
<p>
Reopened tickets: <span class="query_count" title="0 tickets matching status=reopened, max=0, order=id">0</span>
</p>
------------------------------
============================== TicketQuery macro: no results, count 0
Reopened tickets: [[TicketQuery(status=reopened, format=count)]]
------------------------------
<p>
Reopened tickets: <a href="/query?status=reopened&amp;max=0&amp;order=id" title="0 tickets matching status=reopened, max=0, order=id">0</a>
</p>
------------------------------
============================== TicketQuery macro: no results, compact form
Reopened tickets: [[TicketQuery(status=reopened, format=compact)]]
------------------------------
<p>
Reopened tickets: <span class="query_no_results">No results</span>
</p>
------------------------------
============================== TicketQuery macro: one result, list form
New tickets: [[TicketQuery(status=new)]]
------------------------------
<p>
New tickets: </p><div><dl class="wiki compact"><dt><a class="new" href="/ticket/1" title="This is the summary">#1</a></dt><dd>This is the summary</dd></dl></div><p>
</p>
------------------------------
============================== TicketQuery macro: one result, count 1 (raw)
New tickets: [[TicketQuery(status=new, format=rawcount)]]
------------------------------
<p>
New tickets: <span class="query_count" title="1 ticket matching status=new, max=0, order=id">1</span>
</p>
------------------------------
============================== TicketQuery macro: one result, count 1
New tickets: [[TicketQuery(status=new, format=count)]]
------------------------------
<p>
New tickets: <a href="/query?status=new&amp;max=0&amp;order=id" title="1 ticket matching status=new, max=0, order=id">1</a>
</p>
------------------------------
============================== TicketQuery macro: one result, compact form
New tickets: [[TicketQuery(status=new, format=compact)]]
------------------------------
<p>
New tickets: <span><a class="new" href="/ticket/1" title="This is the summary">#1</a></span>
</p>
------------------------------
============================== TicketQuery macro: duplicated fields
New tickets: [[TicketQuery(status=new, format=compact, col=summary|status|status)]]
------------------------------
<p>
New tickets: <span><a class="new" href="/ticket/1" title="This is the summary">#1</a></span>
</p>
------------------------------
============================== TicketQuery macro: duplicated custom fields
New tickets: [[TicketQuery(status=new, format=compact, col=summary|custom1|custom1)]]
------------------------------
<p>
New tickets: <span><a class="new" href="/ticket/1" title="This is the summary">#1</a></span>
</p>
------------------------------
"""

QUERY2_TEST_CASES = u"""
============================== TicketQuery macro: two results, list form
New tickets: [[TicketQuery(status=new, order=reporter)]]
------------------------------
<p>
New tickets: </p><div><dl class="wiki compact"><dt><a class="new" href="/ticket/2" title="This is another summary">#2</a></dt><dd>This is another summary</dd><dt><a class="new" href="/ticket/1" title="This is the summary">#1</a></dt><dd>This is the summary</dd></dl></div><p>
</p>
------------------------------
============================== TicketQuery macro: two results, count 2 (raw)
New tickets: [[TicketQuery(status=new, order=reporter, format=rawcount)]]
------------------------------
<p>
New tickets: <span class="query_count" title="2 tickets matching status=new, max=0, order=reporter">2</span>
</p>
------------------------------
============================== TicketQuery macro: two results, count 2
New tickets: [[TicketQuery(status=new, order=reporter, format=count)]]
------------------------------
<p>
New tickets: <a href="/query?status=new&amp;max=0&amp;order=reporter" title="2 tickets matching status=new, max=0, order=reporter">2</a>
</p>
------------------------------
============================== TicketQuery macro: two results, compact form
New tickets: [[TicketQuery(status=new, order=reporter, format=compact)]]
------------------------------
<p>
New tickets: <span><a class="new" href="/ticket/2" title="This is another summary">#2</a>, <a class="new" href="/ticket/1" title="This is the summary">#1</a></span>
</p>
------------------------------
"""

def query2_setup(tc):
    insert_ticket(tc.env, reporter='santa', summary='This is the summary',
                  status='new')
    insert_ticket(tc.env, reporter='claus', summary='This is another summary',
                  status='new')

def query2_teardown(tc):
    tc.env.reset_db()


COMMENT_TEST_CASES = u"""
============================== comment: link resolver (deprecated)
comment:ticket:1:1 (deprecated)
[comment:ticket:1:1 see above] (deprecated)
comment:ticket:1:description (deprecated)
[comment:ticket:1:description see descr] (deprecated)
comment:ticket:2:1 (deprecated)
comment:ticket:2:3 (deprecated)
comment:ticket:3:1 (deprecated)
comment:tiket:2:1 (deprecated)
comment:ticket:two:1 (deprecated)
comment:ticket:2:1a (deprecated)
comment:ticket:2:one (deprecated)
comment:ticket:1: (deprecated)
comment:ticket::2 (deprecated)
comment:ticket:: (deprecated)
------------------------------
<p>
<a class="new ticket" href="/ticket/1#comment:1" title="Comment 1 for #1: This is the summary for ticket 1 (new)">comment:ticket:1:1</a> (deprecated)
<a class="new ticket" href="/ticket/1#comment:1" title="Comment 1 for #1: This is the summary for ticket 1 (new)">see above</a> (deprecated)
<a class="new ticket" href="/ticket/1#comment:description" title="Description for #1: This is the summary for ticket 1 (new)">comment:ticket:1:description</a> (deprecated)
<a class="new ticket" href="/ticket/1#comment:description" title="Description for #1: This is the summary for ticket 1 (new)">see descr</a> (deprecated)
<a class="ticket" href="/ticket/2#comment:1" title="Comment 1">comment:ticket:2:1</a> (deprecated)
<a class="missing ticket" title="ticket comment does not exist">comment:ticket:2:3</a> (deprecated)
<a class="missing ticket" title="ticket does not exist">comment:ticket:3:1</a> (deprecated)
comment:tiket:2:1 (deprecated)
comment:ticket:two:1 (deprecated)
comment:ticket:2:1a (deprecated)
comment:ticket:2:one (deprecated)
comment:ticket:1: (deprecated)
comment:ticket::2 (deprecated)
comment:ticket:: (deprecated)
</p>
------------------------------
============================== comment: link resolver
comment:1
comment:₁
[comment:1 see above]
comment:description
[comment:description see descr]
comment:
comment:one
comment:1a
------------------------------
<p>
<a class="ticket" href="/ticket/2#comment:1" title="Comment 1">comment:1</a>
<a class="missing ticket" title="ticket comment does not exist">comment:₁</a>
<a class="ticket" href="/ticket/2#comment:1" title="Comment 1">see above</a>
<a class="ticket" href="/ticket/2#comment:description" title="Description">comment:description</a>
<a class="ticket" href="/ticket/2#comment:description" title="Description">see descr</a>
comment:
comment:one
comment:1a
</p>
------------------------------
============================== comment: link resolver with ticket number
comment:1:bug:1
comment:1:issue:1
comment:1:ticket:1
comment:₁:ticket:1
[comment:1:ticket:1 see above]
comment:description:ticket:1
[comment:description:ticket:1 see descr]
comment:1:ticket:2
comment:3:ticket:2
comment:1:ticket:3
comment:2:tiket:1
comment:1:ticket:two
comment:one:ticket:1
comment:1a:ticket:1
comment:ticket:1
comment:2:ticket:
comment::ticket:
------------------------------
<p>
<a class="new ticket" href="/ticket/1#comment:1" title="Comment 1 for #1: This is the summary for ticket 1 (new)">comment:1:bug:1</a>
<a class="new ticket" href="/ticket/1#comment:1" title="Comment 1 for #1: This is the summary for ticket 1 (new)">comment:1:issue:1</a>
<a class="new ticket" href="/ticket/1#comment:1" title="Comment 1 for #1: This is the summary for ticket 1 (new)">comment:1:ticket:1</a>
<a class="missing ticket" title="ticket comment does not exist">comment:₁:ticket:1</a>
<a class="new ticket" href="/ticket/1#comment:1" title="Comment 1 for #1: This is the summary for ticket 1 (new)">see above</a>
<a class="new ticket" href="/ticket/1#comment:description" title="Description for #1: This is the summary for ticket 1 (new)">comment:description:ticket:1</a>
<a class="new ticket" href="/ticket/1#comment:description" title="Description for #1: This is the summary for ticket 1 (new)">see descr</a>
<a class="ticket" href="/ticket/2#comment:1" title="Comment 1">comment:1:ticket:2</a>
<a class="missing ticket" title="ticket comment does not exist">comment:3:ticket:2</a>
<a class="missing ticket" title="ticket does not exist">comment:1:ticket:3</a>
comment:2:tiket:1
comment:1:ticket:two
comment:one:ticket:1
comment:1a:ticket:1
comment:ticket:1
comment:2:ticket:
comment::ticket:
</p>
------------------------------
""" # "

# NOTE: the following test cases:
#
#  comment:2
#  [comment:2 see above]
#
# would trigger an error in the workaround code ../api.py, line 235
# As it's a problem with a temp workaround, I think there's no need
# to fix it for now.

def comment_setup(tc):
    ticket1 = insert_ticket(tc.env, reporter='santa',
                            summary='This is the summary for ticket 1',
                            status='new')
    ticket1.save_changes(comment='This is the comment for ticket 1')
    ticket2 = insert_ticket(tc.env, reporter='claws',
                            summary='This is the summary for ticket 2',
                            status='closed')
    ticket2.save_changes(comment='This is the comment for ticket 2')

def comment_teardown(tc):
    tc.env.reset_db()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(formatter.test_suite(TICKET_TEST_CASES, ticket_setup,
                                       __file__, ticket_teardown))
    suite.addTest(formatter.test_suite(REPORT_TEST_CASES, report_setup,
                                       __file__))
    suite.addTest(formatter.test_suite(MILESTONE_TEST_CASES, milestone_setup,
                                       __file__, milestone_teardown))
    suite.addTest(formatter.test_suite(QUERY_TEST_CASES, ticket_setup,
                                       __file__, ticket_teardown))
    suite.addTest(formatter.test_suite(QUERY2_TEST_CASES, query2_setup,
                                       __file__, query2_teardown))
    suite.addTest(formatter.test_suite(COMMENT_TEST_CASES, comment_setup,
                                       __file__, comment_teardown,
                                       ('ticket', 2)))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.
import textwrap
import unittest
from datetime import datetime

from trac.test import EnvironmentStub, Mock
from trac.tests.contentgen import random_sentence
from trac.ticket.test import insert_ticket
from trac.util.datefmt import utc
from trac.versioncontrol.api import Repository, RepositoryManager
from trac.wiki.tests import formatter
from tracopt.ticket.commit_updater import CommitTicketUpdater


class CommitTicketUpdaterTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(enable=['trac.*',
                                           'tracopt.ticket.commit_updater.*'])
        self.env.config.set('ticket', 'commit_ticket_update_check_perms', False)
        self.repos = Mock(Repository, 'repos1', {'name': 'repos1', 'id': 1},
                          self.env.log, normalize_rev=lambda rev: 1)
        self.updater = CommitTicketUpdater(self.env)

    def tearDown(self):
        self.env.reset_db()

    def _make_tickets(self, num):
        self.tickets = []
        for i in xrange(num):
            ticket = insert_ticket(self.env, reporter='someone',
                                   summary=random_sentence())
            self.tickets.append(ticket)

    def test_changeset_added(self):
        self._make_tickets(1)
        message = 'This is the first comment. Refs #1.'
        chgset = Mock(repos=self.repos, rev=1, message=message, author='joe',
                      date=datetime(2001, 1, 1, 1, 1, 1, 0, utc))
        self.updater.changeset_added(self.repos, chgset)
        changes = self.tickets[0].get_change(cnum=1)
        self.assertEqual(textwrap.dedent("""\
            In [changeset:"1/repos1" 1/repos1]:
            {{{#!CommitTicketReference repository="repos1" revision="1"
            This is the first comment. Refs #1.
            }}}"""), changes['fields']['comment']['new'])

    def test_changeset_added_multiline_comment(self):
        self._make_tickets(1)
        message = ("This is a multiline comment.\n\n"
                   "It is multiline.\n\n"
                   "Refs #1.")
        chgset = Mock(repos=self.repos, rev=1, message=message, author='joe',
                      date=datetime(2001, 1, 1, 1, 1, 1, 0, utc))
        self.updater.changeset_added(self.repos, chgset)
        changes = self.tickets[0].get_change(cnum=1)
        self.assertEqual(textwrap.dedent("""\
            In [changeset:"1/repos1" 1/repos1]:
            {{{#!CommitTicketReference repository="repos1" revision="1"
            This is a multiline comment.

            It is multiline.

            Refs #1.
            }}}"""), changes['fields']['comment']['new'])

    def test_changeset_modified(self):
        self._make_tickets(2)
        message = 'This is the first comment. Refs #1.'
        old_chgset = Mock(repos=self.repos, rev=1,
                          message=message, author='joe',
                          date=datetime(2001, 1, 1, 1, 1, 1, 0, utc))
        message = 'This is the first comment after an edit. Refs #1, #2.'
        new_chgset = Mock(repos=self.repos, rev=1,
                          message=message, author='joe',
                          date=datetime(2001, 1, 2, 1, 1, 1, 0, utc))
        self.updater.changeset_added(self.repos, old_chgset)
        self.updater.changeset_modified(self.repos, new_chgset, old_chgset)
        changes = self.tickets[0].get_change(cnum=1)
        self.assertEqual(textwrap.dedent("""\
            In [changeset:"1/repos1" 1/repos1]:
            {{{#!CommitTicketReference repository="repos1" revision="1"
            This is the first comment. Refs #1.
            }}}"""), changes['fields']['comment']['new'])
        changes = self.tickets[1].get_change(cnum=1)
        self.assertEqual(textwrap.dedent("""\
            In [changeset:"1/repos1" 1/repos1]:
            {{{#!CommitTicketReference repository="repos1" revision="1"
            This is the first comment after an edit. Refs #1, #2.
            }}}"""), changes['fields']['comment']['new'])

    def test_commands_refs(self):
        commands ={(1,): 'Refs #1', (2,): 'refs #2',
                   (3,): 'refs ticket:3#comment:1',
                   (4,5): 'refs ticket:4#comment:description and '
                          'ticket:5#comment:1'}
        self._make_tickets(5)
        rev = 0

        for tkts, cmd in commands.items():
            rev += 1
            message = "This is the first comment. %s." % cmd
            chgset = Mock(repos=self.repos, rev=rev,
                          message=message, author='joe',
                          date=datetime(2001, 1, 1, 1, 1, 1, 0, utc))

            self.updater.changeset_added(self.repos, chgset)
            comment = self.updater.make_ticket_comment(self.repos, chgset)

            for tkt in tkts:
                change = self.tickets[tkt-1].get_change(cnum=1)
                self.assertEqual(comment, change['fields']['comment']['new'])


def macro_setup(tc):
    tc.env = EnvironmentStub(enable=('trac.*',
                                     'tracopt.ticket.commit_updater.*',))
    insert_ticket(tc.env, summary='the summary', status='new')
    def _get_repository(reponame):
        return Mock(get_changeset=_get_changeset, resource=None)
    def _get_changeset(rev=None):
        return Mock(message="the message. refs #1.  ", rev=rev)
    setattr(RepositoryManager(tc.env), 'get_repository', _get_repository)


COMMIT_TICKET_REF_MACRO_TEST_CASES = u"""\
============================== No arguments
[[CommitTicketReference]]
------------------------------
<p>
</p><div class="message"><p>
the message. refs <a class="new ticket" href="/ticket/1" title="#1: the summary (new)">#1</a>.  <br />
</p>
</div><p>
</p>
------------------------------
"""


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(CommitTicketUpdaterTestCase))
    suite.addTest(formatter.test_suite(COMMIT_TICKET_REF_MACRO_TEST_CASES,
                                       macro_setup, __file__,
                                       context=('ticket', 1)))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2007-2023 Edgewall Software
# Copyright (C) 2007 Eli Carter <retracile@gmail.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/.

from trac.core import Component, implements
from trac.ticket.api import ITicketActionController
from trac.ticket.default_workflow import ConfigurableTicketWorkflow
from trac.ticket.model import Priority
from trac.util.html import tag

revision = "$Rev: 17656 $"
url = "$URL: https://svn.edgewall.org/repos/trac/branches/1.4-stable/sample-plugins/workflow/VoteOperation.py $"


class VoteOperation(Component):
    """Provides a simple vote feature.

    This is a sample action controller illustrating how to create
    additional ''operations''.

    Add the `vote` operation to a workflow action, for example:
    {{{#!ini
    vote = new -> new
    vote.operations = vote
    }}}

    Don't forget to add `VoteOperation` to the `workflow` option
    in the `[ticket]` section of TracIni. When added to the default
    value of `workflow`, the line will look like this:
    {{{#!ini
    workflow = ConfigurableTicketWorkflow,VoteOperation
    }}}
    """

    implements(ITicketActionController)

    vote_options = ('for', 'against')
    id_for_action = 'action_%s_vote'

    def get_ticket_actions(self, req, ticket):
        controller = ConfigurableTicketWorkflow(self.env)
        return controller.get_actions_by_operation_for_req(req, ticket, 'vote')

    def get_all_status(self):
        return []

    def render_ticket_action_control(self, req, ticket, action):
        id = self.id_for_action % action
        selected_value = req.args.get(id, self.vote_options[0])
        render_control = tag.select(
            [tag.option(x, selected=(x == selected_value or None))
             for x in self.vote_options], name=id, id=id)
        return ("vote", render_control,
                "Vote on the issue, raising or lowering its priority")

    def get_ticket_changes(self, req, ticket, action):
        id = self.id_for_action % action
        selected = req.args.get(id, self.vote_options[0])
        priorities = list(Priority.select(self.env))
        name_by_val = {int(p.value): p.name for p in priorities}
        ticket_priority_name = ticket._old.get('priority', ticket['priority'])
        ticket_priority = \
            [p for p in priorities if p.name == ticket_priority_name][0]
        if selected == self.vote_options[0]:
            max_val = max(name_by_val)
            new_val = min(max_val, int(ticket_priority.value) + 1)
        else:
            min_val = min(name_by_val)
            new_val = max(min_val, int(ticket_priority.value) - 1)

        return {'priority': name_by_val[new_val]}

    def apply_action_side_effects(self, req, ticket, action):
        pass
